/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.calc;

import bardp.gui.FrameGui;
import bardp.raw.RawSparseZYX;
import bardp.protocol.protocolTexture.TextureControllerSub;
import bardp.roi.controller.RoiModel;
import bardp.utility.MessageGui;
import bardp.utility.ThreadTask;
import java.util.logging.Level;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class RoiKeepOneLargest extends ThreadTask {

    public RoiKeepOneLargest(
            final RawSparseZYX rawZYXSparse,
            final RoiModel roiModel
    ) {
        this.rawZYXSparse = rawZYXSparse;
        this.roiModel = roiModel;
    }

    @Override
    public void run() {

        nz = rawZYXSparse.nz;
        ny = rawZYXSparse.ny;
        nx = rawZYXSparse.nx;

        // construct short[] cluster with default value = 1
        // short _clusterId = (short) 1; // value 1 to first research in clusterIdMap
        // already 1 by construct: cf. TexMainSub:288
        this.clusterIdMap = rawZYXSparse.convZYXInDenseFormatShortWithValue((int) 1);

        // default clustering to initial ClusterIdMap
        this.clusterId = runCluster_fromClusterIDMap_toClusterIdMap("PicROISlip:: clustering");

        // construct ROI from ClusterIdMap
        updateROI_fromClusterIdMap("PicROISlip:: create ROIs");
    }

     @Override
    public void taskCanceled() {
        MessageGui messagGui = new MessageGui(
                "Canceled operation",
                "Not supported yet.",
                MessageGui.WARNING_MESSAGE
        );
    }
    
    @Override
    public void taskEnd() {
    }

    /**
     * first set of id cluster is 2
     *
     * @param log
     * @return clusterId is last (clusterId)
     */
    private int runCluster_fromClusterIDMap_toClusterIdMap(
            final String log
    ) {

        //<editor-fold defaultstate="collapsed" desc="define variables">
        int cell;
        boolean neighbourFound;
        int _neighbourId;

        short[][] matCube;
        int i;
        int matCubeLength;
        int[] neighbours;
        int zyx;

        int znYnX, znYnX_ynX;
        int nYnX = ny * nx;
        int iz, iy, ix;
        int zNeighbour, yNeighbour, xNeighbour;
        //</editor-fold>

        //TODO: optimise with "ParrallelArray" : http://blog.paumard.org/2011/10/26/parallel-arrays-1/
        /**
         * first id value is 2 in clusterIdMap
         */
        int _clusterId = 2;

        for (iz = 0; iz < nz; iz++) {
            znYnX = iz * nYnX;

            //<editor-fold defaultstate="collapsed" desc="neighbour matrix">
            if (iz == 0) {
                matCube = NEIGHBOURS2D;
            } else {
                matCube = NEIGHBOURS3D;
            }
            matCubeLength = matCube[0].length;
            neighbours = new int[matCubeLength];
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="clusterIdMap[]: clustering algorithm">
            for (iy = 0; iy < ny; iy++) {
                znYnX_ynX = znYnX + iy * nx;

                for (ix = 0; ix < nx; ix++) {
                    zyx = znYnX_ynX + ix;

                    if (this.clusterIdMap[zyx] > 0) {
                        // research of neighbourgs
                        _neighbourId = Short.MAX_VALUE;
                        for (i = 0; i < neighbours.length; i++) {
                            neighbours[i] = 0;
                        }
                        neighbourFound = false;

                        for (i = 0; i < matCubeLength; i++) {

                            zNeighbour = iz + matCube[2][i];
                            yNeighbour = iy + matCube[1][i];
                            xNeighbour = ix + matCube[0][i];

                            if ((zNeighbour >= 0) && (zNeighbour < nz)
                                    && (yNeighbour >= 0) && (yNeighbour < ny)
                                    && (xNeighbour >= 0) && (xNeighbour < nx)) {

                                cell = this.clusterIdMap[zNeighbour * nYnX
                                        + yNeighbour * nx
                                        + xNeighbour];
                                if (cell > 1) { // 0
                                    neighbourFound = true;
                                    neighbours[i] = cell;
                                    if (cell < _neighbourId) {
                                        _neighbourId = cell;
                                    }
                                }
                            }
                        }

                        if (neighbourFound) {
                            this.clusterIdMap[zyx] = _neighbourId;

                            for (i = 0; i < matCubeLength; i++) {
                                if ((neighbours[i] > 0) && (neighbours[i] != _neighbourId)) {
                                    changeId(neighbours[i], _neighbourId);
                                }
                            }
                        } else {
                            this.clusterIdMap[zyx] = _clusterId;
                            _clusterId++;
                        }
                    }
                }
            }
            //</editor-fold>
        }

        _clusterId--; // delete id by construct

        FrameGui.LOGGER.log(
                Level.INFO,
                "PicROISlip::runCluster [  {0}  ]\nnbClusters : {1}\n",
                new Object[]{log, _clusterId - 1}
        );
        return _clusterId;
    }

    /**
     * replace all of old id by new id in clusterIdMap
     *
     * @param oldID old id
     * @param newID new id
     * @return
     */
    private void changeId(
            final int oldID,
            final int newID
    ) {
        for (int i = nz * ny * nx - 1; i >= 0; i--) {
            if (clusterIdMap[i] == oldID) {
                clusterIdMap[i] = newID;
            }
        }
    }

    private void updateROI_fromClusterIdMap(
            final String log
    ) {
        double[] numberOfVoxelsByCluster = getNumberOfVoxelsByCluster();
        //<editor-fold defaultstate="collapsed" desc="clusters information">
        roiModel.texControllerSub.informationClusterStr = "";
        String msg = "";
        for (int i = 0; i < numberOfVoxelsByCluster.length; i++) {
            msg += "RoiKeepOneLargest::updateROI_fromClusterIdMap::numberOfVoxelsByCluster [ "+ i +" ] = "+numberOfVoxelsByCluster[i]+"\n";
            if (numberOfVoxelsByCluster[i] > 0) {
                roiModel.texControllerSub.informationClusterStr += numberOfVoxelsByCluster[i] + " ";
            }
        }
        FrameGui.LOGGER.log(
                    Level.INFO,
                    msg
        );
        if (roiModel.texControllerSub.informationClusterStr.length() > 0) {
            roiModel.texControllerSub.informationClusterStr = "Cluster(s) of " + roiModel.texControllerSub.informationClusterStr + " voxels is(are) detected. "
                    + "If more than one cluster the texture calculation is only performed on the largest cluster.";
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="iMaxCluster = find index of max cluster">
        double nbMaxCluster = 0;
        int iMaxCluster = -1;
        for (int i = 0; i < numberOfVoxelsByCluster.length; i++) {
            if (numberOfVoxelsByCluster[i] > nbMaxCluster) {
                nbMaxCluster = numberOfVoxelsByCluster[i];
                iMaxCluster = i;
            }
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="keep iMaxCluster Roi and empty others clusters">
        if (iMaxCluster != -1) {
            float roiVoxMin = nz > 1 ? TextureControllerSub.ROI_VOX_MIN_3D : TextureControllerSub.ROI_VOX_MIN_2D;
            if (numberOfVoxelsByCluster[iMaxCluster] < roiVoxMin) {
                roiModel.texControllerSub.informationClusterStr += " (size too small < "+roiVoxMin+"vx)";
            } else {
                roiModel.texControllerSub.informationClusterStr += " (size ok >= "+roiVoxMin+"vx)";
            }

            // empty other cluster than keeproiIndex
            int cell;
            short ONE = (short) 1;
            short ZERO = (short) 0;
            for (int i = nz * ny * nx - 1; i >= 0; i--) {
                cell = this.clusterIdMap[i];
                clusterIdMap[i] = (cell == iMaxCluster) ? ONE : ZERO;
            }

            rawZYXSparse.deleteAllCoor();
            rawZYXSparse.update(clusterIdMap);
        }
        //</editor-fold>
    }

    /**
     * count the number of voxels / cluster
     *
     * @param nbClusterLength number of clusters
     * @return the number of voxels / cluster
     */
    private double[] getNumberOfVoxelsByCluster() {
        double[] tabNbCluster = new double[(int)this.clusterId + 1];
        int cell;

        for (int i = nz * ny * nx - 1; i >= 0; i--) {
            cell = this.clusterIdMap[i];
            if (cell > 0) {
                tabNbCluster[(int)cell]++;
            }
        }
        return tabNbCluster;
    }

    private final RoiModel roiModel;
    private final RawSparseZYX rawZYXSparse;
    private int nz, ny, nx;

   
    /**
     * number of clusters
     */
    private int clusterId = 0;

    /**
     * id cluster array
     */
    private int[] clusterIdMap = null;

    /**
     * kernel of neighbours search in 2D
     */
    private final static short[][] NEIGHBOURS2D = {
        {-1, -1, 0, +1}, // x
        {0, -1, -1, -1}, // y
        {0, 0, 0, 0} // z
    };

    /**
     * kernel of neighbours search in 3D
     */
    private final static short[][] NEIGHBOURS3D = {
        {-1, -1, 0, +1, +1, +1, -1, 0, -1, 0, -1, 0}, // x
        {0, -1, -1, -1, 0, +1, -1, -1, 0, 0, +1, +1}, // y
        {0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1} // z
    };
}