/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.histo;

import bardp.gui.FrameGui;
import bardp.img.picChart.PicHistogramGUI;
import bardp.roi.controller.RoiModel;
import bardp.session.patient.study.series.SeriesModel;
import bardp.utility.LIFExFonts;
import bardp.utility.MessageGui;
import java.awt.Color;
import java.util.ArrayList;
import java.util.logging.Level;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYBarRenderer;
import org.jfree.chart.title.LegendTitle;
import org.jfree.chart.title.TextTitle;
import org.jfree.chart.title.Title;
import org.jfree.data.statistics.HistogramDataset;
import org.jfree.data.statistics.HistogramType;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class HistogramController implements Runnable {

    public HistogramController(
            final RoiModel roiModel,
            final SeriesModel series,
            final double[] rawZYX,
            final double nbGreyLevels,
            final double intensityResamplingMinBound,
            final double intensityResamplingMaxBound,
            final boolean isHISTO_Show,
            final boolean isHISTO_Save,
            final MessageGui progressGUI
    ) {
        this.roiModel = roiModel;
        this.series = series;
        this.rawZYX = rawZYX;
        this.nbGreyLevels = nbGreyLevels;
        this.intensityResamplingMinBound = intensityResamplingMinBound;
        this.intensityResamplingMaxBound = intensityResamplingMaxBound;
        this.isHISTO_Show = isHISTO_Show;
        this.isHISTO_Save = isHISTO_Save;
        this.messageGUI = progressGUI;
    }

    @Override
    public void run() {
        HistogramDataset histogramDataset = getHistogram();
        if (isHISTO_Show) {
            setHistogramVisible(histogramDataset);
        }
        if (isHISTO_Save) {
            saveHistogramFile(histogramDataset);
        }
        
        runSkewnessKurtosis(histogramDataset);
        messageGUI.incProgress();

        runEnergy(histogramDataset);
        messageGUI.incProgress();

        runEntropy(histogramDataset);
        messageGUI.incProgress();
    }

    /**
     * calculation of Mean Max Skewness Kurtosis
     */
    private void runSkewnessKurtosis(
            final HistogramDataset histogramDataset
    ) {
        final long start_time = System.currentTimeMillis();

        FrameGui.LOGGER.log(Level.INFO, "DISCRETIZED_Mean DISCRETIZED_Max DISCRETIZED_Skewness DISCRETIZED_Kurtosis Processing ... [ start ]");
        int i;
        double sumN, sumD;
       
        final int iSerie = 0;
        final int count = histogramDataset.getItemCount(iSerie);
        
        //<editor-fold defaultstate="collapsed" desc="counts wanted are Y values of histogram">
        double ySum = 0;
        double y[] = new double[count];
        for (i = 0; i < count; i++) {
            y[i] = histogramDataset.getYValue(iSerie, i); // count values
            ySum += y[i];
        }
        double yMean = ySum / (double) count;
        //</editor-fold>
        
        //<editor-fold defaultstate="collapsed" desc="skewness">
        // https://en.wikipedia.org/wiki/Skewness
        {
            sumN = 0d;
            sumD = 0d;
            for (i = 0; i < count; i++) {
                sumN = sumN + Math.pow(y[i] - yMean, 3);
                sumD = sumD + Math.pow(y[i] - yMean, 2);
            }
            sumN /= (double) count;
            sumD /= (double) count;
            skewness = sumN / Math.pow(sumD, 1.5d);
            FrameGui.LOGGER.log(Level.INFO, "DISCRETIZED_HISTO_Skewness={0}", skewness);
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="kurtosis">
        // https://en.wikipedia.org/wiki/Kurtosis
        {
            sumN = 0d;
            sumD = 0d;
            for (i = 0; i < count; i++) {
                sumN = sumN + Math.pow(y[i] - yMean, 4);
                sumD = sumD + Math.pow(y[i] - yMean, 2);
            }
            sumN /= (double) count;
            sumD /= (double) count;
            kurtosis = sumN / Math.pow(sumD, 2);
            FrameGui.LOGGER.log(Level.INFO, "DISCRETIZED_HISTO_Kurtosis={0}", kurtosis);
        }
        
        //<editor-fold defaultstate="collapsed" desc="ExcessKurtosis">
        // HISTO_kurtosis is corrected by a Fisher correction of -3 to center kurtosis on 0 for normal distributions
        // IBSI protocol, p.33
        excessKurtosis = kurtosis - 3d;
        FrameGui.LOGGER.log(Level.INFO, "DISCRETIZED_HISTO_ExcessKurtosis={0}", excessKurtosis);
        //</editor-fold>

        final long end_time = System.currentTimeMillis();
        FrameGui.LOGGER.log(
                Level.INFO,
                "Time calculation of DISCRETIZED_Mean, DISCRETIZED_Max, DISCRETIZED_Skewness and DISCRETIZED_Kurtosis proceding: {0} sec",
                ((end_time - start_time) / 1_000d)
        );
    }

    /**
     * get dataset of histogram
     * <br>with nbGreyLevels,
     * <br>with intensityResamplingMinBound,
     * <br>with intensityResamplingMaxBound);
     *
     * @return dataset
     */
    private HistogramDataset getHistogram() {
        HistogramDataset dataset = new HistogramDataset();
        dataset.setType(HistogramType.FREQUENCY);
        dataset.addSeries(
                roiModel.title,
                rawZYX,
                (int) nbGreyLevels,
                intensityResamplingMinBound,
                intensityResamplingMaxBound
        );
        return dataset;
    }

    private void saveHistogramFile(
            final HistogramDataset histogramDataset
    ) {
        HistogramIO_xls histogramIO_xls = new HistogramIO_xls(
                histogramDataset,
                roiModel
        );
        histogramIO_xls.run();
    }

    /**
     * create GUI of Histogram
     */
    private void setHistogramVisible(
            final HistogramDataset histogramDataset
    ) {

        //<editor-fold defaultstate="collapsed" desc="PicHistogramGUI">
        PicHistogramGUI pHistoGUI = new PicHistogramGUI();
        pHistoGUI.jFreeChart = ChartFactory.createHistogram("Histogram",
                series.seriesUnit.unitY.getLongDescription(), "number #",
                histogramDataset, PlotOrientation.VERTICAL, true, true, false);
        pHistoGUI.chartPanel = new ChartPanel(pHistoGUI.jFreeChart);
        pHistoGUI.getContentPane().add(pHistoGUI.chartPanel, null);

        LegendTitle legendTitle = pHistoGUI.jFreeChart.getLegend(); // against bug of SubTitle

        final XYPlot plot = (XYPlot) pHistoGUI.jFreeChart.getPlot();
        plot.getDomainAxis().setTickLabelFont(LIFExFonts.FONT_SMALL); // X
        plot.getDomainAxis().setLabelFont(LIFExFonts.FONT_SMALL);
        plot.getRangeAxis().setTickLabelFont(LIFExFonts.FONT_SMALL); // Y
        plot.getRangeAxis().setLabelFont(LIFExFonts.FONT_SMALL);

        pHistoGUI.jFreeChart.setBackgroundPaint(bardp.ressources.Res.COLOR_BACKGROUND1);
        pHistoGUI.jFreeChart.getTitle().setFont(LIFExFonts.FONT_MEDIUM);
        pHistoGUI.jFreeChart.getLegend().setItemFont(LIFExFonts.FONT_SMALL);
        //</editor-fold>
        //<editor-fold defaultstate="collapsed" desc="setSubtitles(seriesDescription)">
        ArrayList<Title> subTitles = new ArrayList<>();
        subTitles.add(new TextTitle(series.seriesDescription));
        pHistoGUI.jFreeChart.setSubtitles(subTitles);
        pHistoGUI.jFreeChart.addLegend(legendTitle); // bug SubTitle
        //</editor-fold>
        //<editor-fold defaultstate="collapsed" desc="setColor(), setVisible()">
        XYBarRenderer renderer = (XYBarRenderer) pHistoGUI.jFreeChart.getXYPlot().getRenderer();
        int iSeries = 0;
        Color color = new Color(
                roiModel.color.getRed(),
                roiModel.color.getGreen(),
                roiModel.color.getBlue(),
                128 // transparence de moitié
        );
        renderer.setSeriesPaint(iSeries, color);
        pHistoGUI.pack();
        pHistoGUI.setVisible(true);
        FrameGui.LOGGER.log(Level.INFO, "DISCRETIZED_Histogram Processing ... [ end ]");
        //</editor-fold>
    }

    /**
     * calculation of Energy
     */
    private void runEnergy(
            final HistogramDataset histogramDataset
    ) {
        energy = 0d;

        long start_time = System.currentTimeMillis();
        FrameGui.LOGGER.log(Level.INFO, "\nDISCRETIZED_HISTO_Energy Processing ... [ start ]");

        //<editor-fold defaultstate="collapsed" desc="engergy">
        try {
            int iSerie = 0;

            // get value and sum
            double sum = 0;
            int countNbBins = histogramDataset.getItemCount(iSerie);
            double histo[] = new double[countNbBins];
            for (int i = 0; i < countNbBins; i++) {
                histo[i] = histogramDataset.getYValue(iSerie, i);
                sum += histo[i];
            }
            FrameGui.LOGGER.log(
                    Level.INFO,
                    "\nDISCRETIZED_HISTO_EnergycountNbBins = {0}\n  sum = {1}",
                    new Object[]{countNbBins, sum}
            );

            // get proba[]
            double[] proba = new double[countNbBins];
            for (int i = 0; i < countNbBins; i++) {
                proba[i] = histo[i] / sum;
                //FrameGui.LOGGER.log(Level.INFO, "   proba(" + i + ")=" + proba[i] + "\n");
            }

            for (int i = 0; i < countNbBins; i++) {
                energy += proba[i] * proba[i];
            }
            //</editor-fold>

        } catch (java.lang.Exception Ex) {
            FrameGui.LOGGER.log(Level.INFO, "\nDISCRETIZED_HISTO_Energy Processing ... {0}", Ex);
        }

        long end_time = System.currentTimeMillis();
        long difference = end_time - start_time;
        FrameGui.LOGGER.log(Level.INFO, "\nDISCRETIZED_HISTO_Energy: Time calculation proceding: {0} sec", (difference / 1_000d));

        FrameGui.LOGGER.log(Level.INFO, toString());
    }

    /**
     * calculation of Entropy
     */
    private void runEntropy(
            final HistogramDataset histogramDataset
    ) {

        long start_time = System.currentTimeMillis();
        FrameGui.LOGGER.log(Level.INFO, "\nDISCRETIZED_HISTO_Entropy Processing ... [ start ]");

        //<editor-fold defaultstate="collapsed" desc="entropy">
        try {
            double epsilon = Math.ulp(1d);
            double log2 = Math.log(2);

            // get value and sum
            double sum = 0;
            int iSerie = 0;
            int countNbBins = histogramDataset.getItemCount(iSerie);
            double histo[] = new double[countNbBins];
            for (int i = 0; i < countNbBins; i++) {
                histo[i] = histogramDataset.getYValue(iSerie, i);
                histo[i] += epsilon; // against log(0)
                sum += histo[i];
            }
            FrameGui.LOGGER.log(
                    Level.INFO,
                    "\n  countNbBins = {0}\n  sum = {1}",
                    new Object[]{countNbBins, sum}
            );

            // get proba[]
            double[] proba = new double[countNbBins];
            for (int i = 0; i < countNbBins; i++) {
                proba[i] = histo[i] / sum;
            }

            double _entropy_log10 = 0d;
            double _entropy_log2 = 0d;
            for (int i = 0; i < countNbBins; i++) {
                _entropy_log10 += proba[i] * Math.log10(proba[i]);
                _entropy_log2 += proba[i] * Math.log(proba[i]) / log2; // conversion log base e in log base 2
            }
            entropy_log10 = -_entropy_log10;
            entropy_log2 = -_entropy_log2;

        } catch (java.lang.Exception Ex) {
            FrameGui.LOGGER.log(Level.INFO, "\nDISCRETIZED_HISTO_Entropy Processing ... {0}", Ex);
            entropy_log10 = 0d;
            entropy_log2 = 0d;
        }
        //</editor-fold>
        //</editor-fold>

        long end_time = System.currentTimeMillis();
        long difference = end_time - start_time;
        FrameGui.LOGGER.log(Level.INFO, "\nDISCRETIZED_HISTO_Entropy: Time calculation proceding: {0} sec", (difference / 1_000d));

        FrameGui.LOGGER.log(Level.INFO, toString());
    }

    @Override
    public String toString() {
        return "\nFirst Order (Histogram):"
                + "\n DISCRETIZED_HISTO_Skewness:\t" + getSkewness()
                + "\n DISCRETIZED_HISTO_Kurtosis:\t" + getKurtosis()
                + "\n DISCRETIZED_HISTO_Entropy:\t" + getEntropy_log10()
                + "\n DISCRETIZED_HISTO_Entropy_log2:\t" + getEntropy_log2()
                + "\n DISCRETIZED_HISTO_Energy:\t\t" + getEnergy();
    }

    /**
     * LIFEx::getEnergy
     * <br>is a measure of the sum of the squares of each intensity value.
     * <br>
     * <br>This is a measure of the heterogeneity of the image array, where a greater uniformity implies a greater
     * heterogeneity or a greater range of discrete intensity values.
     *
     * @return
     */
    public double getEnergy() {
        return energy;
    }

    public double getEntropy_log10() {
        return entropy_log10;
    }

    public double getEntropy_log2() {
        return entropy_log2;
    }

    public double getKurtosis() {
        return kurtosis;
    }

    public double getExcessKurtosis() {
        return excessKurtosis;
    }

    public double getSkewness() {
        return skewness;
    }

    /**
     *
     */
    private final RoiModel roiModel;

    /**
     * discrete image
     */
    private final double[] rawZYX;

    private final SeriesModel series;
    private final MessageGui messageGUI;
    
    private final double nbGreyLevels;
    private final double intensityResamplingMinBound;
    private final double intensityResamplingMaxBound;
    private final boolean isHISTO_Show;
    private final boolean isHISTO_Save;

    private double skewness;  //french: Asymétrie
    /**
     * kurtosis is calculated as measure of peakedness of the distribution
     */
    private double kurtosis;

    /**
     * excess Kurtosis is corrected by a Fisher correction of -3 to center kurtosis on 0 for normal distributions
     */
    private double excessKurtosis;

    private double entropy_log10;
    private double entropy_log2;
    private double energy;
}
