/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.io;

import bardp.gui.FrameGui;
import bardp.img.controller.FrameController;
import bardp.protocol.protocolTexture.traditional.ConventionalRimModel;
import bardp.protocol.protocolTexture.gui.TextureBoxOption;
import bardp.protocol.protocolTexture.gui.TextureGlobalROIJPanel;
import bardp.protocol.protocolTexture.params.TextureJPanelOutputParameters;
import bardp.roi.controller.RoiController;
import bardp.roi.controller.RoiMain;
import bardp.roi.controller.RoiMainList;
import bardp.session.patient.PatientModel;
import bardp.session.patient.study.series.SeriesModel;
import bardp.utility.MessageGui;
import bardp.utility.Utils;
import bardp.utility.WorkDir;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Calendar;
import java.util.Date;
import java.util.logging.Level;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class TextureIoCsv {

    /**
     *
     * @param texJPanelOption
     * @param roiMain
     * @param it
     * @param iz location of 2D slice, is -1 for 3D volume
     * @param timeFile
     */
    public TextureIoCsv(
            final TextureGlobalROIJPanel texJPanelOption,
            final RoiMain roiMain,
            final int it,
            final int iz,
            final long timeFile
    ) {
        this.texJPanelOption = texJPanelOption;
        this.roiMain = roiMain;
        this.it = it;
        this.iz = iz;
        this.timeFile = timeFile;
    }

    public void run() {

        //<editor-fold defaultstate="collapsed" desc="check texture process">
        boolean found = false;

        final RoiMainList roiMainList;
        if (roiMain == null) {
            roiMainList = RoiController.roiMainList;
        } else {
            roiMainList = new RoiMainList();
            roiMainList.add(roiMain);
        }

        for (RoiMain r : roiMainList) {
            if (r.roiModel.texControllerSub != null) {
                found = true;
            }
        }

        if (!found) {
            MessageGui messageGUI = new MessageGui(
                    "Texture Protocol",
                    "No texture or no ROI found. Create it and try again.",
                    MessageGui.ERROR_MESSAGE
            );
            return;
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="texturePatientFile">
        final String roiDirectory = WorkDir.getROIValueDirectory(FrameController.seriesRef);
        Utils.createDirectory(roiDirectory);

        final PatientModel patient = FrameController.seriesSelected.fromStudy.fromPatient;
        String sortfilename = patient.patientName.trim().toUpperCase()
                + "_" + patient.patientID.trim().toUpperCase()
                + "_Texture";
        sortfilename = sortfilename.replace("\\", "_");
        sortfilename = sortfilename.replaceAll("/", "_");
        sortfilename = sortfilename + "_" + timeFile + ".csv";

        final String strCsvFile = Utils.prepDirectoryAndFile(roiDirectory, sortfilename);
        final File texturePatientCsvFile = new File(strCsvFile);
        createUpdateFileCsv(texturePatientCsvFile);
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="textureSessionFile"> 
        Utils.createDirectory(FrameGui.getDataDirectory());
        final boolean isScript = !texJPanelOption.scriptingCsv.isEmpty();
        final File textureSessionFile = isScript
                ? new File(texJPanelOption.scriptingCsv)
                : new File(FrameGui.getDataDirectory() + File.separator + "TextureSession.csv");
        createUpdateFileCsv(textureSessionFile);
        //</editor-fold>
    }

    private void createUpdateFileCsv(
            final File file
    ) {
        try {
            boolean addTitle = !file.exists();
            try (BufferedWriter writer = new BufferedWriter(new FileWriter(file, !addTitle))) {
                saveInfoToSerie(writer, FrameController.seriesSelected, addTitle);
            }
            //<editor-fold defaultstate="collapsed" desc="log">
            FrameGui.LOGGER.log(Level.INFO, "Save file: {0}", file.getName());
            //</editor-fold>
        } catch (IOException ex) {
            //<editor-fold defaultstate="collapsed" desc="log">
            FrameGui.LOGGER.log(
                    Level.SEVERE,
                    null,
                    ex
            );
            //</editor-fold>
        }
    }

    private void saveInfoToSerie(
            final BufferedWriter writer,
            final SeriesModel series,
            final boolean addTitle
    ) throws IOException {
        String tmp;

        TextureBoxOption texBoxOption = FrameGui.protocolFrame.texBoxOption;
        TextureJPanelOutputParameters texJPanelOutputParameters
                = texBoxOption.textureJPanelOption.texGlobalROIJPanel.texJFrameOutputParameters.texJPanelOutputParameters;

        //<editor-fold defaultstate="collapsed" desc="all titles">
        if (addTitle) {
            writer.write(FrameGui.SOFTNAME + FrameGui.VERSION_SHORT);
            writer.newLine();
            writer.newLine();

            //<editor-fold defaultstate="collapsed" desc="PatientName title & ROI title">
            writer.write("INFO_PatientName");

            writer.write(",");
            writer.write("INFO_PatientID");

            writer.write(",");
            writer.write("INFO_ProcessDateOfTexture");

            writer.write(",");
            writer.write("INFO_SeriesDate");

            writer.write(",");
            writer.write("INFO_Series");

            writer.write(",");
            writer.write("INFO_ActualFrameDuration");

            writer.write(",");
            writer.write("INFO_NameOfRoi");
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="Conventional Indices title">
            writer.write(",");
            tmp = "CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "min ";
            writer.write(tmp);

            writer.write(",");
            tmp = "CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "mean";
            writer.write(tmp);

            writer.write(",");
            tmp = "CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "std";
            writer.write(tmp);

            writer.write(",");
            tmp = "CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "max";
            writer.write(tmp);

            writer.write(",");
            tmp = "CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Q1";
            writer.write(tmp);

            writer.write(",");
            tmp = "CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Q2";
            writer.write(tmp);

            writer.write(",");
            tmp = "CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Q3";
            writer.write(tmp);

            writer.write(",");
            tmp = "CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Skewness";
            writer.write(tmp);
            writer.write(",");
            tmp = "CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Kurtosis";
            writer.write(tmp);
            writer.write(",");
            tmp = "CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "ExcessKurtosis";
            writer.write(tmp);

            writer.write(",");
            writer.write("CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "peakSphere0.5mL:discretized volume sought");
            writer.write(",");
            writer.write("CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "peakSphere0.5mL(value only for PET or NM)");

            writer.write(",");
            writer.write("CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "peakSphere1mL:discretized volume sought");
            writer.write(",");
            writer.write("CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "peakSphere1mL(value only for PET or NM)");

            writer.write(",");
            writer.write("CONVENTIONAL_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "calciumAgatstonScore[onlyForCT]");
            
            writer.write(",");
            writer.write("CONVENTIONAL_TLG(mL)[onlyForPETorNM]");

            if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                writer.write(",");
                writer.write("CONVENTIONAL_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "min");

                writer.write(",");
                writer.write("CONVENTIONAL_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "mean");

                writer.write(",");
                writer.write("CONVENTIONAL_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "stdev");

                writer.write(",");
                writer.write("CONVENTIONAL_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "max");

                writer.write(",");
                writer.write("CONVENTIONAL_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Volume(mL)");

                writer.write(",");
                writer.write("CONVENTIONAL_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Volume(vx)");

                writer.write(",");
                writer.write("CONVENTIONAL_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "sum");
            }

            writer.write(",");
            tmp = "DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "min ";
            writer.write(tmp);

            writer.write(",");
            tmp = "DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "mean";
            writer.write(tmp);

            writer.write(",");
            tmp = "DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "std";
            writer.write(tmp);

            writer.write(",");
            tmp = "DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "max";
            writer.write(tmp);

            writer.write(",");
            tmp = "DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Q1";
            writer.write(tmp);

            writer.write(",");
            tmp = "DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Q2";
            writer.write(tmp);

            writer.write(",");
            tmp = "DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Q3";
            writer.write(tmp);

            writer.write(",");
            tmp = "DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Skewness";
            writer.write(tmp);
            writer.write(",");
            tmp = "DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Kurtosis";
            writer.write(tmp);
            writer.write(",");
            tmp = "DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "ExcessKurtosis";
            writer.write(tmp);

            writer.write(",");
            writer.write("DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "peakSphere0.5mL:discretized volume sought");
            writer.write(",");
            writer.write("DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "peakSphere0.5mL(value only for PET or NM)");

            writer.write(",");
            writer.write("DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "peakSphere1mL:discretized volume sought");
            writer.write(",");
            writer.write("DISCRETIZED_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "peakSphere1mL(value only for PET or NM)");

            writer.write(",");
            writer.write("DISCRETIZED_TLG(mL)[onlyForPETorNM]");

            if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                writer.write(",");
                writer.write("DISCRETIZED_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "min");

                writer.write(",");
                writer.write("DISCRETIZED_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "mean");

                writer.write(",");
                writer.write("DISCRETIZED_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "stdev");

                writer.write(",");
                writer.write("DISCRETIZED_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "max");

                //  writer.write(",");
                //  writer.write("DISCRETIZED_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Volume(mL)");
                //  writer.write(",");
                //  writer.write("DISCRETIZED_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "Volume(vx)");
                writer.write(",");
                writer.write("DISCRETIZED_RIM_" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + "sum");
            }

            //</editor-fold>
            //<editor-fold defaultstate="collapsed" desc="HistogramSeries title">
            writer.write(",");
            writer.write("DISCRETIZED_HISTO_Skewness");

            writer.write(",");
            writer.write("DISCRETIZED_HISTO_Kurtosis");

            writer.write(",");
            writer.write("DISCRETIZED_HISTO_ExcessKurtosis");

            writer.write(",");
            writer.write("DISCRETIZED_HISTO_Entropy_log10");

            writer.write(",");
            writer.write("DISCRETIZED_HISTO_Entropy_log2");

            writer.write(",");
            writer.write("DISCRETIZED_HISTO_Energy[=Uniformity]");
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="Shape title">
            writer.write(",");
            writer.write("SHAPE_Volume(mL)");

            writer.write(",");
            writer.write("SHAPE_Volume(vx)");

            writer.write(",");
            writer.write("SHAPE_Sphericity[onlyFor3DROI])");

            writer.write(",");
            writer.write("SHAPE_Surface(mm2)[onlyFor3DROI]");

            writer.write(",");
            writer.write("SHAPE_Compacity[onlyFor3DROI]");
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="Texture parameters title + size error">
            writer.write(",");
            writer.write("PARAMS_DistanceOfNeighbours");

            writer.write(",");
            writer.write("PARAMS_NumberOfGreyLevels");

            writer.write(",");
            writer.write("PARAMS_BinSize");

            writer.write(",");
            writer.write("PARAMS_IntensityResampling");

            writer.write(",");
            writer.write("PARAMS_BoundsRangeOfValueAfterDiscretisation(" + series.seriesUnit.unitY.getShortDescriptionWithoutHash() + ")");

            writer.write(",");
            writer.write("PARAMS_ZSpatialResampling");

            writer.write(",");
            writer.write("PARAMS_YSpatialResampling");

            writer.write(",");
            writer.write("PARAMS_XSpatialResampling");

            writer.write(",");
            writer.write("CHECK_Cluster(s)ToSmall");
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="GLCM matrix title">
            writer.write(",");
            writer.write("GLCM_Homogeneity[=InverseDifference]"); // bold

            writer.write(",");
            writer.write("GLCM_Energy[=AngularSecondMoment]");

            writer.write(",");
            writer.write("GLCM_Contrast[=Variance]");

            writer.write(",");
            writer.write("GLCM_Correlation");

            writer.write(",");
            writer.write("GLCM_Entropy_log10");

            writer.write(",");
            writer.write("GLCM_Entropy_log2[=JointEntropy]");

            writer.write(",");
            writer.write("GLCM_Dissimilarity");
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="GLRLM matrix title">
            writer.write(",");
            writer.write("GLRLM_SRE"); // bold

            writer.write(",");
            writer.write("GLRLM_LRE"); // bold

            writer.write(",");
            writer.write("GLRLM_LGRE");

            writer.write(",");
            writer.write("GLRLM_HGRE");

            writer.write(",");
            writer.write("GLRLM_SRLGE");

            writer.write(",");
            writer.write("GLRLM_SRHGE");

            writer.write(",");
            writer.write("GLRLM_LRLGE");

            writer.write(",");
            writer.write("GLRLM_LRHGE");

            writer.write(",");
            writer.write("GLRLM_GLNU");

            writer.write(",");
            writer.write("GLRLM_RLNU");

            writer.write(",");
            writer.write("GLRLM_RP");
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="NGLDM matrix title">
            writer.write(",");
            writer.write("NGLDM_Coarseness");

            writer.write(",");
            writer.write("NGLDM_Contrast");

            writer.write(",");
            writer.write("NGLDM_Busyness");
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="GLZLM matrix title">
            writer.write(",");
            writer.write("GLZLM_SZE");

            writer.write(",");
            writer.write("GLZLM_LZE");

            writer.write(",");
            writer.write("GLZLM_LGZE"); // bold

            writer.write(",");
            writer.write("GLZLM_HGZE"); // bold

            writer.write(",");
            writer.write("GLZLM_SZLGE");

            writer.write(",");
            writer.write("GLZLM_SZHGE");

            writer.write(",");
            writer.write("GLZLM_LZLGE");

            writer.write(",");
            writer.write("GLZLM_LZHGE");

            writer.write(",");
            writer.write("GLZLM_GLNU");

            writer.write(",");
            writer.write("GLZLM_ZLNU");

            writer.write(",");
            writer.write("GLZLM_ZP");
            //</editor-fold>

            writer.write(",");  // z position
            writer.write("TimePosition");

            writer.write(",");  // z location
            writer.write("zLocation[onlyFor2DROI]");

            writer.newLine();
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="all values">
        RoiMainList roiMainList;
        if (roiMain == null) {
            roiMainList = RoiController.roiMainList;
        } else {
            roiMainList = new RoiMainList();
            roiMainList.add(roiMain);
        }
        for (RoiMain r : roiMainList) {
            if (r.roiModel.boxOptionMain.jPanelMenu.jCheckBoxEyes.isSelected()) {

                //<editor-fold defaultstate="collapsed" desc="patient + ROI identity">
                writer.write(Utils.prepContentToCsv(series.fromStudy.fromPatient.patientName));

                writer.write(",");
                writer.write(Utils.prepContentToCsv(series.fromStudy.fromPatient.patientID));

                writer.write(",");
                Calendar c = Calendar.getInstance(); // date du jour
                Date d = c.getTime();
                writer.write(Utils.prepContentToCsv(d.toString()));

                writer.write(",");
                if (series.fromStudy != null && series.fromStudy.date != null) {
                    writer.write(Utils.prepContentToCsv(series.fromStudy.date.toString()));
                }

                writer.write(",");
                writer.write(Utils.prepContentToCsv(series.seriesDescription + "_" + series.modality));

                writer.write(",");

                writer.write(Utils.prepContentToCsv(series.actualFrameDuration / 1_000f / 60f + " min"));

                writer.write(",");
                writer.write("'" + Utils.prepContentToCsv(r.roiModel.title));

                //</editor-fold>
                //<editor-fold defaultstate="collapsed" desc="Conventional Indices values">
                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getMinValue();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getMeanValue();
                    writer.write(String.valueOf(value));
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getStdValue();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getMaxValue();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getPercentile25Value();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getPercentile50Value();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getPercentile75Value();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getSkewness();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getKurtosis();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getExcessKurtosis();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVpeakSphere05mL.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getPeak05mLDiscretizedVolumeSought() / 1_000d; // have mL
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVpeakSphere05mL.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getPeak05mLValue();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVpeakSphere1mL.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getPeak1mLDiscretizedVolumeSought() / 1_000d; // have mL
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }
                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVpeakSphere1mL.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getPeak1mLValue();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }
                
                writer.write(",");
                try {
                    double value = r.roiModel.texControllerSub.calciumQuantitationSub.getTotalCalciumScore();
                    writer.write(String.valueOf(value));                // CALCIUM_AGATSTON_SCORE

                } catch (NullPointerException ex) {
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_TLGmL.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getTlg();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {

                    writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texControllerSub.conventionalRim.getRimResults()) {
                                value += rimContent.min + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }

                    writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texControllerSub.conventionalRim.getRimResults()) {
                                value += rimContent.mean + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }

                    writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texControllerSub.conventionalRim.getRimResults()) {
                                value += rimContent.stdev + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }

                    writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texControllerSub.conventionalRim.getRimResults()) {
                                value += rimContent.max + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }

                    writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texControllerSub.conventionalRim.getRimResults()) {
                                value += rimContent.cm3 + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }

                    writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texControllerSub.conventionalRim.getRimResults()) {
                                value += rimContent.n + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }

                    writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texControllerSub.conventionalRim.getRimResults()) {
                                value += rimContent.sum + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }
                }


                
                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getMinValue();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getMeanValue();
                    writer.write(String.valueOf(value));
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getStdValue();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getMaxValue();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getPercentile25Value();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getPercentile50Value();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getPercentile75Value();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getSkewness();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getKurtosis();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVmin_mean_max.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getExcessKurtosis();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVpeakSphere05mL.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getPeak05mLDiscretizedVolumeSought() / 1_000d; // have mL
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVpeakSphere05mL.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getPeak05mLValue();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVpeakSphere1mL.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getPeak1mLDiscretizedVolumeSought() / 1_000d; // have mL
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }
                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_SUVpeakSphere1mL.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getPeak1mLValue();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxConventional_TLGmL.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.discretizedTraditional.traditionalController.getTlg();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {

                    writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texControllerSub.discretizedRim.getRimResults()) {
                                value += rimContent.min + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }

                    writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texControllerSub.discretizedRim.getRimResults()) {
                                value += rimContent.mean + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }

                    writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texControllerSub.discretizedRim.getRimResults()) {
                                value += rimContent.stdev + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }

                    writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texControllerSub.discretizedRim.getRimResults()) {
                                value += rimContent.max + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }

                    /*  writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texMainSub.discretizedRim.getRimResults()) {
                                value += rimContent.cm3 + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }*/

                     /*    writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texMainSub.discretizedRim.getRimResults()) {
                                value += rimContent.n + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }*/
                    writer.write(",");
                    if (texJPanelOutputParameters.jCheckBoxConventional_RIMmin_mean_max.isSelected()) {
                        try {
                            String value = "";
                            for (ConventionalRimModel rimContent : r.roiModel.texControllerSub.discretizedRim.getRimResults()) {
                                value += rimContent.sum + "|";
                            }
                            writer.write(value);
                        } catch (NullPointerException ex) {
                        }
                    }
                }
                //</editor-fold>

             
                //<editor-fold defaultstate="collapsed" desc="HistogramSeries values">
                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxHistogram_Skewness.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.histogramController.getSkewness();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxHistogram_Kurtosis.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.histogramController.getKurtosis();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxHistogram_Kurtosis.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.histogramController.getExcessKurtosis();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxHistogram_Entropy_log10.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.histogramController.getEntropy_log10();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxHistogram_Entropy_log2.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.histogramController.getEntropy_log2();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxHistogram_Energy.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.histogramController.getEnergy();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }
                //</editor-fold>

                //<editor-fold defaultstate="collapsed" desc="Shape values">
                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxShape_Volume_mL.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getVolume_mL();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxShape_Volume_vx.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getVolume_vx();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxShape_Sphericity.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.texSphericity.getSphericity();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxShape_Compacity.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.texSphericity.getSurfaceArea();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }

                writer.write(",");
                if (texJPanelOutputParameters.jCheckBoxShape_Compacity.isSelected()) {
                    try {
                        double value = r.roiModel.texControllerSub.texSphericity.getCompacity();
                        writer.write(String.valueOf(value));
                    } catch (NullPointerException ex) {
                    }
                }
                //</editor-fold>

                //<editor-fold defaultstate="collapsed" desc="Texture parameters values">
                writer.write(","); // Distance of Neighbour
                try {
                    writer.write(String.valueOf(texJPanelOutputParameters.jComboBoxDistance.getSelectedIndex() + 1));
                } catch (NullPointerException ex) {
                }

                writer.write(",");  // Number of Grey Levels
                try {
                    double value = r.roiModel.texControllerSub.intensityDiscretizationNbGreyLevels;
                    writer.write(String.valueOf(value));
                } catch (NullPointerException ex) {
                }

                writer.write(",");  // Bin size
                try {
                    double value = r.roiModel.texControllerSub.intensityDiscretizationBinSize;
                    writer.write(String.valueOf(value));
                } catch (NullPointerException ex) {
                }

                writer.write(",");  // Intensity Rescaling
                if (texBoxOption.textureJPanelOption.texGlobalROIJPanel.jRadioButtonAbsoluteMinMax.isSelected()) {
                    writer.write("Absolute bounds: "
                            + texBoxOption.textureJPanelOption.texGlobalROIJPanel.getIntensityRescalingMinBound()
                            + " <-> "
                            + texBoxOption.textureJPanelOption.texGlobalROIJPanel.getIntensityRescalingMaxBound()
                    );
                } else if (texBoxOption.textureJPanelOption.texGlobalROIJPanel.jRadioButtonRelativeMinMax.isSelected()) {
                    writer.write("Relative ROI: min max: "
                            + r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getMinValue()
                            + " <-> "
                            + r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getMaxValue());
                } else if (texBoxOption.textureJPanelOption.texGlobalROIJPanel.jRadioButtonRelativeSd.isSelected()) {
                    writer.write("Relative ROI: mean+-3sd: "
                            + (r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getMeanValue()
                            - 3 * r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getStdValue())
                            + " <-> "
                            + (r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getMeanValue()
                            + 3 * r.roiModel.texControllerSub.conventionalTraditional.traditionalController.getStdValue())
                    );
                }

                writer.write(",");  // Bounds range of value after discretisation
                String minValue = String.valueOf(r.roiModel.texControllerSub.rawZYXCropDiscretMinValue); // deja en SUV si PET
                String maxValue = String.valueOf(r.roiModel.texControllerSub.rawZYXCropDiscretMaxValue);
                writer.write(minValue + " <-> " + maxValue);

                writer.write(",");  // Z Spatial resampling
                try {
                    double value = r.roiModel.texControllerSub.spatialResampling.z;
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // Y Spatial resampling
                try {
                    double value = r.roiModel.texControllerSub.spatialResampling.y;
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // X Spatial resampling
                try {
                    double value = r.roiModel.texControllerSub.spatialResampling.x;
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // error ROI or multi clusters
                writer.write(r.roiModel.texControllerSub.informationClusterStr);
                //</editor-fold>

                //<editor-fold defaultstate="collapsed" desc="GLCM matrix values">
                writer.write(",");  // Homogeneity
                try {
                    double value = r.roiModel.texControllerSub.texGLCM.getHomogeneity();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(","); // Energy
                try {
                    double value = r.roiModel.texControllerSub.texGLCM.getEnergy();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // Contrast
                try {
                    double value = r.roiModel.texControllerSub.texGLCM.getContrast();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // Correlation
                try {
                    double value = r.roiModel.texControllerSub.texGLCM.getCorrelation();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // Entropy_log10
                try {
                    double value = r.roiModel.texControllerSub.texGLCM.getEntropy_log10();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(","); // Entropy_log2
                try {
                    double value = r.roiModel.texControllerSub.texGLCM.getEntropy_log2();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // Dissimilarity
                try {
                    double value = r.roiModel.texControllerSub.texGLCM.getDissimilarity();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }
                //</editor-fold>

                //<editor-fold defaultstate="collapsed" desc="GLRLM matrix values">
                writer.write(",");  // SRE
                try {
                    double value = r.roiModel.texControllerSub.texGLRLM.getSRE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // LRE
                try {
                    double value = r.roiModel.texControllerSub.texGLRLM.getLRE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // LGRE
                try {
                    double value = r.roiModel.texControllerSub.texGLRLM.getLGRE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(","); // HGRE
                try {
                    double value = r.roiModel.texControllerSub.texGLRLM.getHGRE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // SRLGE
                try {
                    double value = r.roiModel.texControllerSub.texGLRLM.getSRLGE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // SRHGE
                try {
                    double value = r.roiModel.texControllerSub.texGLRLM.getSRHGE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // LRLGE
                try {
                    double value = r.roiModel.texControllerSub.texGLRLM.getLRLGE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // LRHGE
                try {
                    double value = r.roiModel.texControllerSub.texGLRLM.getLRHGE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // GLNU
                try {
                    double value = r.roiModel.texControllerSub.texGLRLM.getGLNU();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // RLNU
                try {
                    double value = r.roiModel.texControllerSub.texGLRLM.getRLNU();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // RP
                try {
                    double value = r.roiModel.texControllerSub.texGLRLM.getRP();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }
                //</editor-fold>

                //<editor-fold defaultstate="collapsed" desc="NGLDM matrix values">
                writer.write(",");  // Coarseness
                try {
                    double value = r.roiModel.texControllerSub.texNGLDM.getCoarseness();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // Contrast
                try {
                    double value = r.roiModel.texControllerSub.texNGLDM.getContrast();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // Busyness
                try {
                    double value = r.roiModel.texControllerSub.texNGLDM.getBusyness();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }
                //</editor-fold>

                //<editor-fold defaultstate="collapsed" desc="GLZLM matrix values">
                writer.write(",");  // SZE
                try {
                    double value = r.roiModel.texControllerSub.texGLZLM.getSZE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // LZE
                try {
                    double value = r.roiModel.texControllerSub.texGLZLM.getLZE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // LGZE
                try {
                    double value = r.roiModel.texControllerSub.texGLZLM.getLGZE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // HGZE
                try {
                    double value = r.roiModel.texControllerSub.texGLZLM.getHGZE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // SZLGE
                try {
                    double value = r.roiModel.texControllerSub.texGLZLM.getSZLGE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // SZHGE
                try {
                    double value = r.roiModel.texControllerSub.texGLZLM.getSZHGE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // LZLGE
                try {
                    double value = r.roiModel.texControllerSub.texGLZLM.getLZLGE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // LZHGE
                try {
                    double value = r.roiModel.texControllerSub.texGLZLM.getLZHGE();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // GLNU
                try {
                    double value = r.roiModel.texControllerSub.texGLZLM.getGLNU();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // ZLNU
                try {
                    double value = r.roiModel.texControllerSub.texGLZLM.getZLNU();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }

                writer.write(",");  // ZP
                try {
                    double value = r.roiModel.texControllerSub.texGLZLM.getZP();
                    writer.write(String.valueOf(value));

                } catch (NullPointerException ex) {
                }
                //</editor-fold>

                writer.write(",");  // time position
                writer.write(String.valueOf(it));

                writer.write(",");  // z location
                writer.write(String.valueOf(iz));

                writer.newLine();
            }
        }
        //</editor-fold>
        //</editor-fold>  
    }

    final private TextureGlobalROIJPanel texJPanelOption;
    final private RoiMain roiMain;
    final private int it;
    final private int iz;
    final private long timeFile;
}
