/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.shape;

import bardp.gui.FrameGui;
import bardp.raw.RawSparseZYX;
import bardp.utility.MessageGui;
import bardp.utility.Utils;
import java.util.logging.Level;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class CalcSphericity implements Runnable {

    /**
     *
     * @param maskOnSeriesRaw
     * @param progressGUI
     */
    public CalcSphericity(
            final RawSparseZYX maskOnSeriesRaw,
            final MessageGui progressGUI
    ) {
        this.maskOnSeriesRaw = maskOnSeriesRaw;
        this.messageGUI = progressGUI;
    }

    @Override
    public void run() {
        long start_time = System.currentTimeMillis();
        FrameGui.LOGGER.log(Level.INFO, "Sphericity processing ... [ start ]");
        
        /*
        // before LIFEx5.86
        try {
            surface = (new InterfaceSurfaceArea()).run(maskOnSeriesRaw, 1); // 1 by construct
        } catch (Exception ex) {
            FrameGui.LOGGER.log(Level.INFO, "Surface error: {0}", ex.toString());
        }*/
        
        surfaceArea = (new Shape()).calculateSurfaceArea(maskOnSeriesRaw);
                
        final double volume = maskOnSeriesRaw.getSizeOfCoorList()
                        * maskOnSeriesRaw.iopIppSp.spModel.getVoxelSpacing();

        sphericity = computeSphericity(volume, surfaceArea);
        compacity = computeCompacity(volume, surfaceArea);
        
        messageGUI.incProgress();

        //<editor-fold defaultstate="collapsed" desc="log">
        FrameGui.LOGGER.log(
                Level.INFO,
                "Sphericity processing ... [ end ]\n"
                + "- time calculation of sphericity proceding: {0} sec\n"
                + "- surface    = {1} mm2\n"
                + "- volume     = {2} mm3\n"
                + "- sphericity = {3}\n"
                + "- compacity  = {4}\n",
                new Object[]{
                    (System.currentTimeMillis() - start_time) / 1_000d,
                    surfaceArea,
                    volume,
                    sphericity,
                    compacity
                }
        );
        //</editor-fold>
    }

    private double computeSphericity(
            final double volume,
            final double surfaceArea
    ) {
        // normalization constant such that sphere has sphericity equal to 1 
        // double sphere = 36d * Math.PI * volume * volume / (surface * surface * surface); // LIFEx<5.31

        // https://en.wikipedia.org/wiki/Sphericity
        double sphere = StrictMath.pow(Math.PI, 1d / 3d) * StrictMath.pow(6d * volume, 2d / 3d) / surfaceArea;
        /*if (sphere > 1d) {
            sphere = 1d;
        }*/
        return sphere;
    }

    private double computeCompacity(
            final double volume,
            final double surface
    ) {
        return volume / surface;
    }

    @Override
    public String toString() {
        return "\nSphericity:\t\t" + Utils.d2s(getSphericity(), 1)
                + "\nCompacity:\t\t" + Utils.d2s(getCompacity(), 1);
    }

    public double getSphericity() {
        return sphericity;
    }

    public double getSurfaceArea() {
        return surfaceArea;
    }

    public double getCompacity() {
        return compacity;
    }

    public void setIsTooSmall() {
        sphericity = Double.NaN;
        compacity = Double.NaN;
        surfaceArea = Double.NaN;
    }

    private double sphericity;
    private double compacity;
    private double surfaceArea;

    /**
     * ROI
     */
    private final RawSparseZYX maskOnSeriesRaw;
    private final MessageGui messageGUI;
}
