/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package bardp.protocol.protocolTexture.shape;

import bardp.session.patient.study.series.image.SpModel;
import inra.ijpb.geometry.Vector3D;
import inra.ijpb.measure.region3d.GeometryUtils;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: Utility functions for classes that compute 3D intrinsic volumes measures.</p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author dlegland
 * @see inra.ijpb.measure.IntrinsicVolumes3D
 * @version 6.0
 */
public class IntrinsicVolumes3DUtilsToDelete
{
    /**
     * Returns an array with seven values corresponding the unique direction
     * vectors obtained with 13 directions.
     * 
     * @param spModel
     * @return an array of seven values, corresponding to the three isothetic,
     *         three plane diagonal, and one cube diagonal directions.
     */
    public static final double[] directionWeights3d13(
            final SpModel spModel
    ) {
        // extract resolution as individual variables
        double dx = spModel.getXSpacing();
        double dy = spModel.getYSpacing();
        double dz = spModel.getZSpacing();
        
        // allocate memory for resulting array
        double[] weights = new double[7];
    
        // In case of cubic grid, use pre-computed weights
        if (dx == dy && dx == dz) 
        {
            // in case of cubic voxels, uses pre-computed weights
            weights[0] = 0.04577789120476 * 2;  // Ox
            weights[1] = 0.04577789120476 * 2;  // Oy
            weights[2] = 0.04577789120476 * 2;  // Oz
            weights[3] = 0.03698062787608 * 2;  // Oxy
            weights[4] = 0.03698062787608 * 2;  // Oxz
            weights[5] = 0.03698062787608 * 2;  // Oyz
            weights[6] = 0.03519563978232 * 2;  // Oxyz
            return weights;
        }

        // Create a set of reference vectors, named after their contribution to
        // each direction: 'P' stands for positive, 'N' stands for negative, 
        // and 'Z' stands for Zero. 
        // Hence, vector vZPN has x-coordinate equal to zero, y-coordinate 
        // equal to +dy, and z-coordinate equal to -dz. 
            
        // direction vectors pointing below the OXY plane
        Vector3D vPNN = new Vector3D( dx, -dy, -dz).normalize(); 
        Vector3D vPZN = new Vector3D( dx,   0, -dz).normalize(); 
        Vector3D vNPN = new Vector3D(-dx,  dy, -dz).normalize(); 
        Vector3D vZPN = new Vector3D(  0,  dy, -dz).normalize(); 
        Vector3D vPPN = new Vector3D( dx,  dy, -dz).normalize(); 
        
        // direction vectors pointing belonging to the OXY plane
        Vector3D vPNZ = new Vector3D( dx, -dy,   0).normalize();
        Vector3D vPZZ = new Vector3D( dx,   0,   0).normalize();
        Vector3D vNPZ = new Vector3D(-dx,  dy,   0).normalize();
        Vector3D vZPZ = new Vector3D(  0,  dy,   0).normalize();
        Vector3D vPPZ = new Vector3D( dx,  dy,   0).normalize();
        
        // direction vectors pointing above the OXY plane
        Vector3D vNNP = new Vector3D(-dx, -dy,  dz).normalize(); 
        Vector3D vZNP = new Vector3D(  0, -dy,  dz).normalize(); 
        Vector3D vPNP = new Vector3D( dx, -dy,  dz).normalize(); 
        Vector3D vNZP = new Vector3D(-dx,   0,  dz).normalize(); 
        Vector3D vZZP = new Vector3D(  0,   0,  dz).normalize(); 
        Vector3D vPZP = new Vector3D( dx,   0,  dz).normalize(); 
        Vector3D vNPP = new Vector3D(-dx,  dy,  dz).normalize(); 
        Vector3D vZPP = new Vector3D(  0,  dy,  dz).normalize(); 
        Vector3D vPPP = new Vector3D( dx,  dy,  dz).normalize(); 
    
        Vector3D[] neighbors;
        
        double piAndHalf = 2 * Math.PI; // orig
        
        // Spherical cap type 1, direction [1 0 0]
        neighbors = new Vector3D[]{vPNN, vPNZ, vPNP, vPZP, vPPP, vPPZ, vPPN, vPZN};
        weights[0] = GeometryUtils.sphericalVoronoiDomainArea(vPZZ, neighbors) / piAndHalf;
        
        // Spherical cap type 1, direction [0 1 0]
        neighbors = new Vector3D[]{vPPZ, vPPP, vZPP, vNPP, vNPZ, vNPN, vZPN, vPPN};
        weights[1] = GeometryUtils.sphericalVoronoiDomainArea(vZPZ, neighbors) / piAndHalf;
    
        // Spherical cap type 1, direction [0 0 1]
        neighbors = new Vector3D[]{vPZP, vPPP, vZPP, vNPP, vNZP, vNNP, vZNP, vPNP};
        weights[2] = GeometryUtils.sphericalVoronoiDomainArea(vZZP, neighbors) / piAndHalf;
    
        // Spherical cap type 2, direction [1 1 0]
        neighbors = new Vector3D[]{vPZZ, vPPP, vZPZ, vPPN};
        weights[3] = GeometryUtils.sphericalVoronoiDomainArea(vPPZ, neighbors) / piAndHalf;
    
        // Spherical cap type 2, direction [1 0 1]
        neighbors = new Vector3D[]{vPZZ, vPPP, vZZP, vPNP};
        weights[4] = GeometryUtils.sphericalVoronoiDomainArea(vPZP, neighbors) / piAndHalf;
    
        // Spherical cap type 2, direction [0 1 1]
        neighbors = new Vector3D[]{vZPZ, vNPP, vZZP, vPPP};
        weights[5] = GeometryUtils.sphericalVoronoiDomainArea(vZPP, neighbors) / piAndHalf;
    
        // Spherical cap type 2, direction [1 0 1]
        neighbors = new Vector3D[]{vPZP, vZZP, vZPP, vZPZ, vPPZ, vPZZ};
        weights[6] = GeometryUtils.sphericalVoronoiDomainArea(vPPP, neighbors) / piAndHalf;
        
        return weights;
    }
}