/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.textureGLCM;

import bardp.gui.FrameGui;
import bardp.img.coor.CoorRange;
import bardp.raw.RawTZYX;
import bardp.raw.RawSparseZYX;
import bardp.utility.MessageGui;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.logging.Level;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class TextureGLCMController implements Runnable {

    /**
     * processing the cooccurrence matrix of gray level
     * <p>
     * @param outputFilename
     * @param isCsvGLCMSave
     * @param dist distance with neighbours
     * @param rawZYXDiscret discretized image (validate only with SUV unit)
     * @param maskOnSeriesRaw
     * @param rangeBounds
     * @param messageGUI
     */
    public TextureGLCMController(
            String outputFilename,
            boolean isCsvGLCMSave,
            int dist,
            RawTZYX rawZYXDiscret,
            RawSparseZYX maskOnSeriesRaw,
            CoorRange rangeBounds,
            MessageGui messageGUI
    ) {
        this.outputFilename = outputFilename;
        this.isCsvGLCMSave = isCsvGLCMSave;
        this.dist = dist;
        this.rawZYXDiscret = rawZYXDiscret;
        this.maskOnSeriesRaw = maskOnSeriesRaw;
        this.rangeBounds = rangeBounds;
        this.messageGUI = messageGUI;
        this.GLCM = null;
    }

    @Override
    public void run() {
        long start_time = System.currentTimeMillis();

        FrameGui.LOGGER.log(Level.INFO, "GLCM Processing ... [ start ]");
        outOfMemory = false;

        int nbDirections;
        int[][] directions;
        
        if (rangeBounds.max.getZ() - rangeBounds.min.getZ() == 0) {
            // 2D AXIAL
            nbDirections = NB_DIRECTIONS_2D;
            directions = DIRECTIONS_2D_OF_AXIAL;
        } else if (rangeBounds.max.getY() - rangeBounds.min.getY() == 0) {
            // 2D CORONAL
            nbDirections = NB_DIRECTIONS_2D;
            directions = DIRECTIONS_2D_OF_CORONAL;
        } else if (rangeBounds.max.getX() - rangeBounds.min.getX() == 0) {
            // 2D SAGITTAL
            nbDirections = NB_DIRECTIONS_2D;
            directions = DIRECTIONS_2D_OF_SAGITTAL;
        } else { 
            // 3D
            nbDirections = NB_DIRECTIONS_3D;
            directions = DIRECTIONS_3D;
        }
  
        // define grey level
        int maxValue = (int) Math.ceil(rawZYXDiscret.maxValue);
        int nbLevel = maxValue + 1;

        //<editor-fold defaultstate="collapsed" desc="GLCM core processing">
        try {
            TextureGLCMSubRoutine texGLCMSubRoutine = new TextureGLCMSubRoutine(
                    maskOnSeriesRaw,
                    rawZYXDiscret,
                    nbDirections, directions,
                    dist,
                    nbLevel
            );
            GLCM = texGLCMSubRoutine.getGLCMDirection();
        } catch (Exception ex) {
            FrameGui.LOGGER.log(Level.SEVERE, null, ex);
        }
        //</editor-fold>
        //</editor-fold>

        double[][] matDirec;
        int r, c;
        double[][] trans_arr = new double[nbLevel][nbLevel];

        for (int iDirection = 0; iDirection < nbDirections; iDirection++) {

            matDirec = GLCM[iDirection];

            if (FrameGui.DEBUG) {
                System.out.println("Av:iDirection=" + iDirection);
                printDebug(matDirec);
            }

            // matrix transpose
            for (r = 0; r < nbLevel; r++) {
                for (c = 0; c < nbLevel; c++) {
                    trans_arr[r][c] = matDirec[c][r];
                }
            }

            // sum of transpose
            for (r = 0; r < nbLevel; r++) {
                for (c = 0; c < nbLevel; c++) {
                    matDirec[r][c] += trans_arr[r][c];
                }
            }
        }

        //</editor-fold>
        //<editor-fold defaultstate="collapsed" desc="GLCM normalisation processing">
        for (int direction = 0; direction < nbDirections; direction++) {
            GLCM[direction] = calculateNormalizeMatrix(GLCM[direction]);
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="save GLCM.csv file">
        if (isCsvGLCMSave) {
            FileOutputStream fileout = null;
            try {
                long time = System.currentTimeMillis();
                fileout = new FileOutputStream(outputFilename + "_GLCM_" + time + ".csv");
                try (BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(fileout))) {

                    writer.write("normalized GLCM");
                    writer.newLine();
                    writer.write("TextureUserGuide.pdf - section 1.1");
                    writer.newLine();
                    writer.newLine();
                    writer.write("number of directions " + nbDirections);
                    writer.newLine();
                    writer.write("search distance of neighbours (vx)" + dist);
                    writer.newLine();
                    for (int iDirec = 0; iDirec < nbDirections; iDirec++) {
                        writer.newLine();
                        writer.write("Direction " + (iDirec + 1) + ":");
                        writer.newLine();
                        writer.write(",Level (p)");
                        writer.newLine();
                        writer.write("Level (q),,");
                        for (int iLevel = 1; iLevel < nbLevel; iLevel++) { // be careful: iLevel start to 1
                            writer.write((iLevel) + ",");
                        }
                        writer.newLine();
                        for (int iLevel = 1; iLevel < nbLevel; iLevel++) { // be careful: iLevel start to 1
                            writer.write("," + (iLevel) + ",");
                            for (int jLevel = 1; jLevel < nbLevel; jLevel++) { // be careful: iLevel start to 1
                                writer.write(Double.toString(GLCM[iDirec][iLevel][jLevel]));
                                writer.write(",");
                            }
                            writer.newLine();
                        }
                    }
                }
                FrameGui.LOGGER.log(
                        Level.INFO,
                        "Save GLCM.csv file: {0}",
                        fileout
                );
                MessageGui messageGUI = new MessageGui(
                        "csv File [ OK ]",
                        "GLCM.csv",
                        MessageGui.OK_MESSAGE
                );
            } catch (IOException ex) {
                FrameGui.LOGGER.log(
                        Level.SEVERE,
                        null,
                        ex
                );
            } finally {
                try {
                    if (fileout != null) {
                        fileout.close();
                    }
                } catch (IOException ex) {
                    FrameGui.LOGGER.log(
                            Level.SEVERE,
                            null, ex
                    );
                }
            }
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="Homogeneity processing">
        homogeneity = 0;
        double[] homogeneityByDirection = new double[nbDirections];
        for (int direction = 0; direction < nbDirections; direction++) {
            homogeneityByDirection[direction] = calculateHomogeneity(GLCM[direction]);
            homogeneity += homogeneityByDirection[direction];
        }
        homogeneity /= (double) nbDirections;
        //</editor-fold>    

        //<editor-fold defaultstate="collapsed" desc="Energy processing">
        energy = 0;
        double[] energyByDirection = new double[nbDirections];
        for (int direction = 0; direction < nbDirections; direction++) {
            energyByDirection[direction] = calculateEnergy(GLCM[direction]);
            energy += energyByDirection[direction];
        }
        energy /= (double) nbDirections;
        //FrameGui.logger.log(Level.INFO, "GLCM Energy = {0}", getEnergy(GLCM.get(0)));
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="Contrast processing">
        contrast = 0;
        double[] contrastByDirection = new double[nbDirections];
        for (int direction = 0; direction < nbDirections; direction++) {
            contrastByDirection[direction] = calculateContrast(GLCM[direction]);
            contrast += contrastByDirection[direction];
        }
        contrast /= (double) nbDirections;
        //FrameGui.logger.log(Level.INFO, "GLCM Contrast = {0}", getContrast(GLCM.get(0)));
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="Correlation processing">
        correlation = 0;
        double[] correlationByDirection = new double[nbDirections];
        for (int direction = 0; direction < nbDirections; direction++) {
            correlationByDirection[direction] = calculateCorrelation(GLCM[direction]);
            correlation += correlationByDirection[direction];
        }
        correlation /= (double) nbDirections;
        //FrameGui.logger.log(Level.INFO, "GLCM Correlation = {0}", calculateCorrelation(GLCM.get(0)));
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="Entropy_log10 processing">
        entropy_log10 = 0;
        double[] entropyByDirection = new double[nbDirections];
        for (int direction = 0; direction < nbDirections; direction++) {
            entropyByDirection[direction] = calculateEntropy_log10(GLCM[direction]);
            entropy_log10 += entropyByDirection[direction];
        }
        entropy_log10 /= (double) nbDirections;
        //FrameGui.logger.log(Level.INFO, "GLCM Entropy_log10 = {0}", calculateEntropy_log10(GLCM.get(0)));
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="Entropy_log2 processing">
        entropy_log2 = 0;
        for (int direction = 0; direction < nbDirections; direction++) {
            entropy_log2 += calculateEntropy_log2(GLCM[direction]);
        }
        entropy_log2 /= (double) nbDirections;
        //FrameGui.logger.log(Level.INFO, "GLCM Entropy_log2 = {0}", calculateEntropy_log2(GLCM.get(0)));
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="Dissimilarity processing">
        dissimilarity = 0;
        double[] dissimilarityByDirection = new double[nbDirections];
        for (int direction = 0; direction < nbDirections; direction++) {
            dissimilarityByDirection[direction] = calculateDissimilarity(GLCM[direction]);
            dissimilarity += dissimilarityByDirection[direction];
        }
        dissimilarity /= (float) nbDirections;
        // FrameGui.logger.log(Level.INFO, "GLCM Dissimilarity = {0}", calculateDissimilarity(GLCM.get(0)));
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="save TI by direction in fileoutTIByDirection.csv">
        if (isCsvGLCMSave) {
            FileOutputStream fileoutTIByDirection = null;
            long time = System.currentTimeMillis();
            try {
                fileoutTIByDirection = new FileOutputStream(outputFilename + "_GLCM_TIDirection_" + time + ".csv");
            } catch (FileNotFoundException ex) {
                FrameGui.LOGGER.log(Level.SEVERE, null, ex);
            }

            try (BufferedWriter writerTIByDirection = new BufferedWriter(new OutputStreamWriter(fileoutTIByDirection))) {
                try {
                    writerTIByDirection.write("TI of normalized GLCM by each direction");
                    writerTIByDirection.newLine();

                    writerTIByDirection.write(
                            ",x dir"
                            + ",y dir"
                            + ",z dir"
                            + ",distance"
                            + ",homogeneity"
                            + ",energy"
                            + ",contrast"
                            + ",correlation"
                            + ",entropy"
                            + ",dissimilarity"
                    );
                    writerTIByDirection.newLine();
                    for (int direction = 0; direction < nbDirections; direction++) {
                        writerTIByDirection.write(
                                "direction " + (direction + 1)
                                + "," + directions[0][direction]
                                + "," + directions[1][direction]
                                + "," + directions[2][direction]
                                + "," + dist
                                + "," + homogeneityByDirection[direction]
                                + "," + energyByDirection[direction]
                                + "," + contrastByDirection[direction]
                                + "," + correlationByDirection[direction]
                                + "," + entropyByDirection[direction]
                                + "," + dissimilarityByDirection[direction]
                        );
                        writerTIByDirection.newLine();
                    }
                    writerTIByDirection.write(
                            "mean"
                            + ","
                            + ","
                            + ","
                            + ","
                            + "," + homogeneity
                            + "," + energy
                            + "," + contrast
                            + "," + correlation
                            + "," + entropy_log10
                            + "," + dissimilarity
                    );

                } catch (IOException ex) {
                    FrameGui.LOGGER.log(
                            Level.SEVERE,
                            null, ex);
                }

            } catch (IOException ex) {
                FrameGui.LOGGER.log(
                        Level.SEVERE,
                        null, ex);
            }

            try {
                fileoutTIByDirection.close();
            } catch (IOException ex) {
                FrameGui.LOGGER.log(
                        Level.SEVERE,
                        null, ex);
            }
        }
        //</editor-fold>

        FrameGui.LOGGER.log(Level.INFO, "GLCM Processing ... [ end ]");
        messageGUI.incProgress();

        long end_time = System.currentTimeMillis();
        long difference = end_time - start_time;
        FrameGui.LOGGER.log(Level.INFO,
                "Time calculation of GLCM proceding: {0} sec",
                (difference / 1_000d)
        );
    }

    /**
     * matrix normalization
     * <p>
     * @param mat
     * @return
     */
    private double[][] calculateNormalizeMatrix(
            final double[][] mat
    ) {
        int nr = mat.length;
        int nc = mat[0].length;
        double[][] _normMatrix = new double[nr][nc];

        int r, c;

        double sum = 0;
        for (r = 0; r < nr; r++) {
            for (c = 0; c < nc; c++) {
                sum += mat[r][c];
            }
        }
        for (r = 0; r < nr; r++) {
            for (c = 0; c < nc; c++) {
                _normMatrix[r][c] = mat[r][c] / sum;
            }
        }

        return _normMatrix;
    }

    /**
     * processing of Entropy in log10
     * <br>with matrix "mat" must be normalized
     * <p>
     * @param mat
     * @return homogeneity, range = [0 1]
     */
    private double calculateEntropy_log10(
            final double[][] mat
    ) {
        double epsilon = Math.ulp(1d);
        double _entropy = 0;
        int r, c;
        int nr, nc;

        nr = mat.length;
        nc = mat[0].length;

        for (r = 0; r < nr; r++) {
            for (c = 0; c < nc; c++) {
                _entropy += mat[r][c] * Math.log10(mat[r][c] + epsilon);
            }
        }
        return -_entropy;
    }

    /**
     * processing of Entropy in log2
     * <br>with matrix "mat" must be normalized
     * <p>
     * @param mat
     * @return homogeneity, range = [0 1]
     */
    private double calculateEntropy_log2(
            final double[][] mat
    ) {

        double epsilon = Math.ulp(1d);
        double log2 = Math.log(2);
        double _entropy = 0;
        for (int r = 0; r < mat.length; r++) {
            for (int c = 0; c < mat[0].length; c++) {
                _entropy += mat[r][c] * Math.log(mat[r][c] + epsilon) / log2;
            }
        }
        return -_entropy;
    }

    /**
     * calculate of Dissimilarity value
     * <br>be careful "mat" matrix must be normalized before
     * <p>
     * @param mat
     * @return homogeneity, range = [0 1]
     */
    private double calculateDissimilarity(
            final double[][] mat
    ) {
        double _dissimilarity = 0;
        int r, c;
        int nr, nc;

        nr = mat.length;
        nc = mat[0].length;

        for (r = 0; r < nr; r++) {
            for (c = 0; c < nc; c++) {
                _dissimilarity += mat[r][c] * Math.abs(r - c);
            }
        }
        return _dissimilarity;
    }

    /**
     * calculate of Homogeneity value
     * <br>be careful "mat" matrix must be normalized before
     * <p>
     * @param mat
     * @return homogeneity, range = [0 1]
     */
    private double calculateHomogeneity(
            final double[][] mat
    ) {
        double _homogeneity = 0;
        int r, c;
        int nr, nc;

        nr = mat.length;
        nc = mat[0].length;

        for (r = 0; r < nr; r++) {
            for (c = 0; c < nc; c++) {
                _homogeneity += mat[r][c] / (1d + Math.abs(r - c));
            }
        }
        return _homogeneity;
    }

    /**
     * calculate of Energy value
     * <br>be careful "mat" matrix must be normalized before
     * <p>
     * @param mat
     * @return energy, range = [0 1] Reference: Haralick RM, Shapiro LG.
     * Computer and Robot Vision: Vol. 1, Addison-Wesley, 1992, p. 460.
     */
    private double calculateEnergy(
            final double[][] mat
    ) {
        double _energy = 0;
        int r, c;
        int nr, nc;

        nr = mat.length;
        nc = mat[0].length;

        for (r = 0; r < nr; r++) {
            for (c = 0; c < nc; c++) {
                _energy += mat[r][c] * mat[r][c];
            }
        }
        return _energy;
    }

    /**
     * calculate of correlation value
     * <br>be careful "mat" matrix must be normalized before
     * <p>
     * @param mat
     * @return energie, Range = [-1 1] References: Haralick RM, Shapiro LG.
     * Computer and Robot Vision: Vol. 1, Addison-Wesley, 1992, p. 460. Bevk M,
     * Kononenko I. A Statistical Approach to Texture Description of Medical
     * Images: A Preliminary Study., The Nineteenth International Conference of
     * Machine Learning, Sydney, 2002.
     * http://www.cse.unsw.edu.au/~icml2002/workshops/MLCV02/MLCV02-Bevk.pdf,
     * p.3.
     */
    private double calculateCorrelation(
            final double[][] mat
    ) {

        int r, c;
        int nr, nc;

        nr = mat.length;
        nc = mat[0].length;

        // Correlation is defined as the covariance(r,c) / S(r)*S(c)
        // where S is the standard deviation.
        //<editor-fold defaultstate="collapsed" desc="mr = Calculate the mean of a pixel value in the row">
        double mr = 0;
        for (c = 0; c < nc; c++) {
            for (r = 0; r < nr; r++) {
                mr += (r + 1) * mat[r][c];
            }
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="sr = Calculate the standard deviation of a pixel value in the row">
        double sr = 0;
        for (c = 0; c < nc; c++) {
            for (r = 0; r < nr; r++) {
                sr += Math.pow((r + 1) - mr, 2) * mat[r][c];
            }
        }
        sr = Math.sqrt(sr);
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="mc = Calculate the mean of pixel value in the column">
        double mc = 0;
        for (c = 0; c < nc; c++) {
            for (r = 0; r < nr; r++) {
                mc += (c + 1) * mat[r][c];
            }
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="sc = Calculate the standard deviation of pixel value in the column">
        double sc = 0;
        for (c = 0; c < nc; c++) {
            for (r = 0; r < nr; r++) {
                sc += Math.pow((c + 1) - mc, 2) * mat[r][c];
            }
        }
        sc = Math.sqrt(sc);
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="Calculate of covariance">
        double cov = 0;
        for (c = 0; c < nc; c++) {
            for (r = 0; r < nr; r++) {
                cov += ((r + 1) - mr) * ((c + 1) - mc) * mat[r][c];
            }
        }
        //</editor-fold>

        //<editor-fold defaultstate="collapsed" desc="Calculate of correlation">
        double _correlation = 0;
        try {
            _correlation = cov / (sr * sc);
        } catch (ArithmeticException ae) {
            FrameGui.LOGGER.log(Level.INFO, "GLCM_Correlation Processing / division by zero ... [ error ]");
        }
        //</editor-fold>
        //</editor-fold>

        return _correlation;
    }

    /**
     * Returns a measure of the intensity contrast between a pixel and its
     * neighbor over the whole image.
     * <br>Range = [0 (size(GLCM,1)-1)^2]
     * <br>Contrast is 0 for a constant image.
     * <br>be careful "mat" matrix must be normalized before
     * <p>
     * @param mat
     * @return contrast
     */
    private double calculateContrast(
            final double[][] mat
    ) {
        double _contrast = 0;
        int r, c;
        int nc, nr;
        nr = mat.length;
        nc = mat[0].length;

        for (r = 0; r < nr; r++) {
            for (c = 0; c < nc; c++) {
                _contrast += Math.abs(r - c) * Math.abs(r - c) * mat[r][c];
            }
        }
        return _contrast;
    }

    @Override
    public String toString() {

        return "\nFirst Order (GLCM matrix):"
                + "\nHomogeneity:\t\t" + getHomogeneity()
                + "\nEnergy:\t\t" + getEnergy()
                + "\nContrast:\t\t" + getContrast()
                + "\nCorrelation:\t\t" + getCorrelation()
                + "\nEntropy:\t\t" + getEntropy_log10()
                + "\nEntropy_log2:\t\t" + getEntropy_log2()
                + "\nDissimilarity:\t\t" + getDissimilarity();
    }

    public double getEnergy() {
        return energy;
    }

    public double getHomogeneity() {
        return homogeneity;
    }

    public double getContrast() {
        return contrast;
    }

    public double getCorrelation() {
        return correlation;
    }

    public double getDissimilarity() {
        return dissimilarity;
    }

    public double getEntropy_log10() {
        return entropy_log10;
    }

    public double getEntropy_log2() {
        return entropy_log2;
    }

    /**
     * print Debug
     */
    private void printDebug(double[][] mat) {
        int n1 = mat.length;
        int n2 = mat[0].length;

        System.out.println("");

        for (int x = 0; x < n1; x++) {
            FrameGui.LOGGER.log(Level.INFO, "");
            for (int y = 0; y < n2; y++) {
                if (mat[x][y] > 0) {
                    System.out.printf(" %.0f", mat[x][y]);
                } else {
                    System.out.printf(" .");
                }
            }
            System.out.println("");
        }
    }

    public void setIsTooSmall() {
        homogeneity = Double.NaN;
        energy = Double.NaN;
        correlation = Double.NaN;
        contrast = Double.NaN;
        entropy_log10 = Double.NaN;
        entropy_log2 = Double.NaN;
        dissimilarity = Double.NaN;
    }

    /**
     * matlab: ex, graycoprops pour vérification
     * <p>
     * >> GLCM = [0 1 2 3;1 1 2 3;1 0 2 0;0 0 0 3]; >> stats = graycoprops(GLCM)
     * <p>
     * stats = Contrast: 2.8947 Correlation: 0.0783 Energy: 0.1191 Homogeneity:
     * 0.5658
     */
    private double[][] matDebug() {

        int nbLevel = 4;
        //  int maxValue = 3;

        double[][] mat = new double[nbLevel][nbLevel];

        mat[0][0] = 0;
        mat[0][1] = 1;
        mat[0][2] = 2;
        mat[0][3] = 3;

        mat[1][0] = 1;
        mat[1][1] = 1;
        mat[1][2] = 2;
        mat[1][3] = 3;

        mat[2][0] = 1;
        mat[2][1] = 0;
        mat[2][2] = 2;
        mat[2][3] = 0;

        mat[3][0] = 0;
        mat[3][1] = 0;
        mat[3][2] = 0;
        mat[3][3] = 3;

        //  print(g.get(0));
        return mat;
    }

    /**
     * true if processing give an outOfMemory
     */
    boolean outOfMemory;

    /**
     * discrete image
     */
    private final RawTZYX rawZYXDiscret;
    private final RawSparseZYX maskOnSeriesRaw;
    private final CoorRange rangeBounds;

    private double[][][] GLCM;

    private final int NB_DIRECTIONS_3D = 13;
    private static final int[][] DIRECTIONS_3D = {
        //1  2   3  4  5   6  7  8   9  10  11  12  13
        {1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, 0}, //x
        {0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0}, //y
        {1, 0, -1, 1, 0, -1, 1, 0, -1, 1, 0, -1, 1} //z
    };

    private final int NB_DIRECTIONS_2D = 4;
    private static final int[][] DIRECTIONS_2D_OF_AXIAL = {
        //1 2  3  4 
        {1, 1, 0, -1}, //x
        {0, 1, 1, 1}, //y
        {0, 0, 0, 0} //z
    };
    
    private static final int[][] DIRECTIONS_2D_OF_CORONAL = {
        //1 2  3  4 
        {1, 1, 0, -1}, //x
        {0, 0, 0, 0}, //
        {0, 1, 1, 1} //z
    };

    private static final int[][] DIRECTIONS_2D_OF_SAGITTAL = {
        //1 2  3  4 
        {0, 0, 0, 0}, //
        {1, 1, 0, -1}, //y
        {0, 1, 1, 1} //z
        
    };

    /**
     * homogeneity
     * <br>Returns a value that measures the closeness of the distribution of
     * elements in the GLCM to the GLCM diagonal.
     * <br>Range = [0 1]
     * <br>Homogeneity is 1 for a diagonal GLCM.
     */
    private double homogeneity;

    /**
     * Returns the sum of squared elements in the GLCM. Range = [0 1] Energy is
     * 1 for a constant image.
     */
    private double energy;

    /**
     * Returns a measure of how correlated a pixel is to its neighbor over the
     * whole image. Range = [-1 1] Correlation is 1 or -1 for a perfectly
     * positively or negatively correlated image. Correlation is NaN for a
     * constant image.
     */
    private double correlation;

    /**
     * Returns a measure of the intensity contrast between a pixel and its
     * neighbor over the whole image. Range = [0 (size(GLCM,1)-1)^2] Contrast is
     * 0 for a constant image.
     */
    private double contrast;

    /**
     *
     */
    private double entropy_log10;
    private double entropy_log2;

    /**
     *
     */
    private double dissimilarity;

    private final MessageGui messageGUI;
    private final String outputFilename;
    private final boolean isCsvGLCMSave;
    private final int dist;

}
