/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.textureGLRLM;

import bardp.gui.FrameGui;
import bardp.img.coor.CoorRange;
import bardp.raw.RawTZYX;
import bardp.raw.RawSparseZYX;
import bardp.utility.MessageGui;
import bardp.utility.ThreadExecution;
import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.logging.Level;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class TextureGLRLM implements Runnable {

    /**
     * procedure of processing the grey level run length matrix (GLRLM)
     * <p>
     * @param outputFilename
     * @param csvGRLMSave
     * @param rawZYXDiscret (unit SUV)
     * @param mask
     * @param rangeBounds bounds of CoorList
     * @param messageGUI
     */
    public TextureGLRLM(
            String outputFilename,
            boolean csvGRLMSave,
            RawTZYX rawZYXDiscret,
            RawSparseZYX mask,
            CoorRange rangeBounds,
            MessageGui messageGUI
    ) {
        this.outputFilename = outputFilename;
        this.csvGRLMSave = csvGRLMSave;
        this.rawZYXDiscret = rawZYXDiscret;
        this.mask = mask;
        this.rangeBounds = rangeBounds;
        this.messageGUI = messageGUI;
    }

    @Override
    public void run() {

        long start_time = System.currentTimeMillis();

        FrameGui.LOGGER.log(Level.INFO, "GLRLM Processing ... [ start ]");
        outOfMemory = false;
        try {
            int nbDirections = rangeBounds.max.getZ() + 1 - rangeBounds.min.getZ() >= 2
                    ? NB_DIRECTIONS_3D
                    : NB_DIRECTIONS_2D;

            //<editor-fold defaultstate="collapsed" desc="Debug: displaying rawZYXDiscret">
            /*FrameGui.LOGGER.log(Level.INFO, "tum_discret_dil:");
            for (int z = rb.min.getFloorZ(); z <= rb.max.getFloorZ(); z++) {
                System.out.println("----------------------------------------------------------------------------");
                System.out.printf("z%2d   ", z - rb.min.getFloorZ() + 1);
                for (int x = rb.min.getFloorX(); x <= rb.max.getFloorX(); x++) {
                    System.out.printf("   x%d", (x - rb.min.getFloorX()) + 1);
                }
                System.out.println("");
                for (int y = rb.max.getFloorY(); y >= rb.min.getFloorY(); y--) {

                    System.out.printf("y%2d |", y - rb.min.getFloorY() + 1);
                    for (int x = rb.max.getFloorX(); x >= rb.min.getFloorX(); x--) {
                        System.out.printf("   %2.0f", rawZYXDiscret.getXYZ(x, y, z));
                    }
                    System.out.println("");
                }
                System.out.print("");
            }*/
            //</editor-fold>
            //<editor-fold defaultstate="collapsed" desc="maxAlignVoxels = longueur maximum d'alignement des voxels (unité : voxel)">
            int maxAlignVoxels = Math.max(rangeBounds.max.getX() - rangeBounds.min.getX() + 1, rangeBounds.max.getY() - rangeBounds.min.getY() + 1);
            maxAlignVoxels = Math.max(maxAlignVoxels, rangeBounds.max.getZ() - rangeBounds.min.getZ() + 1);
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="nbGrayLevels = determination du niveau de gris maximum">
            int maxValue = (int) Math.ceil(rawZYXDiscret.maxValue);
            int nbGreyLevels = maxValue;
            //</editor-fold>

            GLRLM = new int[nbGreyLevels][maxAlignVoxels][nbDirections];

            //<editor-fold defaultstate="collapsed" desc="GLRLM processing with 13 directions">
            ArrayList<Runnable> tasks = new ArrayList<>();
            for (int iDirection = 0; iDirection < nbDirections; iDirection++) {
                TextureGLRLMSubRoutine texGLRLMSub = new TextureGLRLMSubRoutine(GLRLM, rawZYXDiscret, mask,
                        rangeBounds, iDirection);
                tasks.add(texGLRLMSub);
            }
            ThreadExecution.runWithThreadsAndWait(tasks, 2 * 60);

            //</editor-fold> 
            //</editor-fold> 
            //<editor-fold defaultstate="collapsed" desc="Debug: display GLRLM matrix">
            /*     System.out.println("_______________ GLRLM ___________________");
            int[] sum = new int[nbDirections];
            for (int iDirection = 0; iDirection < nbDirections; iDirection++) {
                System.out.println("GLRLM Direction = " + (iDirection + 1));
                for (int iGrayLevel = 0; iGrayLevel < nbGreyLevels; iGrayLevel++) {
                    for (int iAlignVoxel = 0; iAlignVoxel < maxAlignVoxels; iAlignVoxel++) {
                        sum[iDirection] += GLRLM[iGrayLevel][iAlignVoxel][iDirection];
                        System.out.printf(" %3d", GLRLM[iGrayLevel][iAlignVoxel][iDirection]);
                    }
                    System.out.println("");
                }
            }
            System.out.println("Sum=");
            for (int iDirection = 0; iDirection < nbDirections; iDirection++) {
                    System.out.println("sum["+ iDirection + "] = " + sum[iDirection]);
            }*/
            //</editor-fold>
            //<editor-fold defaultstate="collapsed" desc="save GLRLM.csv file">
            if (csvGRLMSave) {
                FileOutputStream fileout = null;
                try {
                    long time = System.currentTimeMillis();
                    fileout = new FileOutputStream(outputFilename + "_GLRLM_" + time + ".csv");
                    try (BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(fileout))) {

                        writer.write("GLRLM");
                        writer.newLine();
                        writer.write("TextureUserGuide.pdf - 3.1p22");
                        writer.newLine();
                        writer.newLine();
                        writer.write("number of directions " + nbDirections);
                        writer.newLine();
                        for (int iDirection = 0; iDirection < nbDirections; iDirection++) {
                            writer.newLine();
                            writer.write("Direction " + (iDirection + 1) + ":");
                            writer.newLine();

                            writer.write("Size run (j)\t");
                            for (int iAlignVoxel = 0; iAlignVoxel < maxAlignVoxels; iAlignVoxel++) {
                                writer.write((iAlignVoxel + 1) + "\t");
                            }
                            writer.newLine();

                            for (int iGrayLevel = 0; iGrayLevel < nbGreyLevels; iGrayLevel++) {
                                writer.write("Intensity (i) " + (iGrayLevel + 1) + "\t");
                                for (int iAlignVoxel = 0; iAlignVoxel < maxAlignVoxels; iAlignVoxel++) {
                                    writer.write(Integer.toString(GLRLM[iGrayLevel][iAlignVoxel][iDirection]) + "\t");
                                }
                                writer.newLine();
                            }
                            writer.newLine();
                        }
                    }
                    FrameGui.LOGGER.log(Level.INFO, "Save GLRLM.csv file: " + fileout);
                    MessageGui messageGUI = new MessageGui(
                            "csv File [ OK ]",
                            "GLRLM.csv",
                            MessageGui.OK_MESSAGE);
                } catch (IOException ex) {
                    FrameGui.LOGGER.log(Level.SEVERE,
                            null,
                            ex);
                } finally {
                    try {
                        if (fileout != null) {
                            fileout.close();
                        }
                    } catch (IOException ex) {
                        FrameGui.LOGGER.log(Level.SEVERE,
                                null, ex);
                    }
                }
            }
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="long[] nbRuns = Number of Runs by direction">
            long[] nbRuns = new long[nbDirections];
            long nbRunsAllDirections = 0;
            long nbVxTotInROIs = 0;

            for (int iDirection = 0; iDirection < nbDirections; iDirection++) {
                for (int iGreyLevel = 0; iGreyLevel < nbGreyLevels; iGreyLevel++) {
                    for (int jAlignVoxels = 0; jAlignVoxels < maxAlignVoxels; jAlignVoxels++) {
                        // nbRuns
                        nbRuns[iDirection] += GLRLM[iGreyLevel][jAlignVoxels][iDirection];

                        // nbVxTotInROI by construct of GLRLM
                        nbVxTotInROIs += (jAlignVoxels + 1) * GLRLM[iGreyLevel][jAlignVoxels][iDirection];
                    }
                }
                nbRunsAllDirections += nbRuns[iDirection];
            }
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="SRE LGRE HGRE SRLGE SRHGE LRLGE LRHGE processing">
            int jAlignVoxels;
            int iGreyLevel;
            int iDirection;

            {
                SRE = 0;
                double SREsum;
                LRE = 0;
                double LREsum;
                LGRE = 0;
                double LGREsum;
                HGRE = 0;
                double HGREsum;
                SRLGE = 0;
                double SRLGEsum;
                SRHGE = 0;
                double SRHGEsum;
                LRLGE = 0;
                double LRLGEsum;
                LRHGE = 0;
                double LRHGEsum;

                double greyLevel2, alignVoxels2;

                for (iDirection = 0; iDirection < nbDirections; iDirection++) {

                    SREsum = 0;
                    LREsum = 0;
                    LGREsum = 0;
                    HGREsum = 0;
                    SRLGEsum = 0;
                    SRHGEsum = 0;
                    LRLGEsum = 0;
                    LRHGEsum = 0;

                    for (iGreyLevel = 0; iGreyLevel < nbGreyLevels; iGreyLevel++) {

                        for (jAlignVoxels = 0; jAlignVoxels < maxAlignVoxels; jAlignVoxels++) {

                            greyLevel2 = (double) ((iGreyLevel + 1) * (iGreyLevel + 1)); // iGrelLevelMin = 0
                            alignVoxels2 = (double) ((jAlignVoxels + 1) * (jAlignVoxels + 1)); // jAlignVoxelsMin = 0

                            SREsum += GLRLM[iGreyLevel][jAlignVoxels][iDirection] / alignVoxels2;
                            LREsum += GLRLM[iGreyLevel][jAlignVoxels][iDirection] * alignVoxels2;
                            LGREsum += GLRLM[iGreyLevel][jAlignVoxels][iDirection] / greyLevel2;
                            HGREsum += GLRLM[iGreyLevel][jAlignVoxels][iDirection] * greyLevel2;
                            SRLGEsum += GLRLM[iGreyLevel][jAlignVoxels][iDirection] / greyLevel2 / alignVoxels2;
                            SRHGEsum += GLRLM[iGreyLevel][jAlignVoxels][iDirection] * greyLevel2 / alignVoxels2;
                            LRLGEsum += GLRLM[iGreyLevel][jAlignVoxels][iDirection] / greyLevel2 * alignVoxels2;
                            LRHGEsum += GLRLM[iGreyLevel][jAlignVoxels][iDirection] * greyLevel2 * alignVoxels2;
                        }
                    }

                    SRE += SREsum / (double) nbRuns[iDirection];
                    LRE += LREsum / (double) nbRuns[iDirection];
                    LGRE += LGREsum / (double) nbRuns[iDirection];
                    HGRE += HGREsum / (double) nbRuns[iDirection];
                    SRLGE += SRLGEsum / (double) nbRuns[iDirection];
                    SRHGE += SRHGEsum / (double) nbRuns[iDirection];
                    LRLGE += LRLGEsum / (double) nbRuns[iDirection];
                    LRHGE += LRHGEsum / (double) nbRuns[iDirection];
                }
                SRE /= nbDirections;
                LRE /= nbDirections;
                LGRE /= nbDirections;
                HGRE /= nbDirections;
                SRLGE /= nbDirections;
                SRHGE /= nbDirections;
                LRLGE /= nbDirections;
                LRHGE /= nbDirections;
            }//</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="GLRLM_GLNU processing">
            {
                GLNU = 0d;

                double sumGreyLevel;
                double sumAlignVoxels;

                for (iDirection = 0; iDirection < nbDirections; iDirection++) {
                    sumGreyLevel = 0d;
                    for (iGreyLevel = 0; iGreyLevel < nbGreyLevels; iGreyLevel++) {
                        sumAlignVoxels = 0d;
                        for (jAlignVoxels = 0; jAlignVoxels < maxAlignVoxels; jAlignVoxels++) {
                            sumAlignVoxels += GLRLM[iGreyLevel][jAlignVoxels][iDirection];
                        }
                        sumGreyLevel += sumAlignVoxels * sumAlignVoxels;
                    }
                    GLNU += sumGreyLevel / (double) nbRuns[iDirection];
                }
                GLNU /= (double) nbDirections;
            }
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="GLRLM_RLNU processing">
            {
                RLNU = 0d;

                double sumGreyLevel;
                double sumAlignVoxels;

                for (iDirection = 0; iDirection < nbDirections; iDirection++) {
                    sumAlignVoxels = 0d;
                    for (jAlignVoxels = 0; jAlignVoxels < maxAlignVoxels; jAlignVoxels++) {
                        sumGreyLevel = 0d;
                        for (iGreyLevel = 0; iGreyLevel < nbGreyLevels; iGreyLevel++) {
                            sumGreyLevel += GLRLM[iGreyLevel][jAlignVoxels][iDirection];
                        }
                        sumAlignVoxels += sumGreyLevel * sumGreyLevel;
                    }
                    RLNU += sumAlignVoxels / (double) nbRuns[iDirection];
                }
                RLNU /= (double) nbDirections;
            }
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="GLRLM_RunPourcentage processing">
            RP = nbRunsAllDirections / (double) nbVxTotInROIs;
            //</editor-fold>

        } catch (OutOfMemoryError ex) {
            outOfMemory = true;
            SRE = 0;
            LRE = 0;
            LGRE = 0;
            HGRE = 0;
            SRLGE = 0;
            SRHGE = 0;
            LRLGE = 0;
            LRHGE = 0;
            GLNU = 0;
            RLNU = 0;
            RP = 0;
            FrameGui.LOGGER.log(Level.SEVERE, "GLRLM Processing ... [  OutOfMemoryError  ]");
        }

        messageGUI.incProgress();
        FrameGui.LOGGER.log(Level.INFO, "GLRLM Processing ... [ end ]");
        long end_time = System.currentTimeMillis();
        long difference = end_time - start_time;
        FrameGui.LOGGER.log(Level.INFO, "Time calculation of GLRLM proceding: {0} sec", (difference / 1_000d));
    }

    @Override
    public String toString() {
        return "\nSecond Order (GLRLM matrix):"
                + "\n  SRE:\t\t" + getSRE()
                + "\n  LRE:\t\t" + getLRE()
                + "\n  LGRE:\t\t" + getLGRE()
                + "\n  HGRE:\t\t" + getHGRE()
                + "\n  SRLGE:\t\t" + getSRLGE()
                + "\n  SRHGE:\t\t" + getSRHGE()
                + "\n  LRLGE:\t\t" + getLRLGE()
                + "\n  LRHGE:\t\t" + getLRHGE()
                + "\n  GLNU:\t\t" + getGLNU()
                + "\n  RLNU:\t\t" + getRLNU()
                + "\n  RP:\t\t" + getRP();
    }

    public double getGLNU() {
        return GLNU;
    }

    public double getHGRE() {
        return HGRE;
    }

    public double getLGRE() {
        return LGRE;
    }

    public double getLRE() {
        return LRE;
    }

    public double getLRHGE() {
        return LRHGE;
    }

    public double getLRLGE() {
        return LRLGE;
    }

    public double getRLNU() {
        return RLNU;
    }

    public double getSRE() {
        return SRE;
    }

    public double getSRHGE() {
        return SRHGE;
    }

    public double getSRLGE() {
        return SRLGE;
    }

    public double getRP() {
        return RP;
    }

    public void setIsTooSmall() {
        SRE = Double.NaN;
        LRE = Double.NaN;
        LGRE = Double.NaN;
        HGRE = Double.NaN;
        SRLGE = Double.NaN;
        SRHGE = Double.NaN;
        LRLGE = Double.NaN;
        LRHGE = Double.NaN;
        GLNU = Double.NaN;
        RLNU = Double.NaN;
        RP = Double.NaN;
    }

    /**
     * discrete image
     */
    private final RawTZYX rawZYXDiscret;
    private final CoorRange rangeBounds;
    private final RawSparseZYX mask;

    private int[][][] GLRLM;

    private final int NB_DIRECTIONS_3D = 13;
    private final int NB_DIRECTIONS_2D = 4;

    /**
     * Short Run Emphasis SRE
     */
    private double SRE;

    /**
     * Long Run Emphasis LRE
     */
    private double LRE;
    /**
     * Low Gray-Level Run Emphasis LGRE
     */
    private double LGRE;
    /**
     * High Gray-Level Run Emphasis HGRE
     */
    private double HGRE;
    /**
     * Short Run Low Gray-Level Emphasis SRLGE
     */
    private double SRLGE;
    /**
     * Short Run High Gray-Level Emphasis SRHGE
     */
    private double SRHGE;
    /**
     * Long Run Low Gray-Level Emphasis LRLGE
     */
    private double LRLGE;
    /**
     * Long Run High Gray-Level Emphasis LRHGE
     */
    private double LRHGE;
    /**
     * Gray-Level Non-Uniformity GLNU
     */
    private double GLNU;
    /**
     * Run Length Non-Uniformity RLNU
     */
    private double RLNU;
    /**
     * Run Percentage
     */
    private double RP;

    private final MessageGui messageGUI;

    /**
     * true if outOfMemory processing
     */
    public boolean outOfMemory;

    private final String outputFilename;
    private final boolean csvGRLMSave;

}
