/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.textureGLZLM;

import bardp.gui.FrameGui;
import bardp.img.coor.CoorRange;
import bardp.raw.RawTZYX;
import bardp.raw.RawSparseZYX;
import bardp.utility.MessageGui;
import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class TextureGLZLM implements Runnable {

    /**
     * procedure of processing the Gray-Level Zone Length Matrix (GLZLM)
     * <p>
     * @param outputFilename
     * @param rawZYXDiscret discrete image (mandatory unit: SUV)
     * @param csvGLZLMSave
     * @param maskRaw
     * @param messageGUI
     * @param rb
     * @param nbCores
     */
    public TextureGLZLM(
            String outputFilename,
            boolean csvGLZLMSave,
            RawTZYX rawZYXDiscret,
            RawSparseZYX maskRaw,
            CoorRange rb,
            MessageGui messageGUI,
            int nbCores
    ) {
        this.outputFilename = outputFilename;
        this.csvGLZLMSave = csvGLZLMSave;
        this.rawZYXDiscret = rawZYXDiscret;
        this.maskRaw = maskRaw;
        this.rb = rb;
        this.messageGUI = messageGUI;
        this.nbCores = nbCores;
    }

    @Override
    public void run() {
        long start_time = System.currentTimeMillis();

        FrameGui.LOGGER.log(Level.INFO, "GLZLM Processing ... [ start ]");
        outOfMemory = false;
        try {

            //<editor-fold defaultstate="collapsed" desc="Debug : affichage rawZYXDiscret">
            /*   FrameGui.logger.log(Level.INFO, "tum_discret_dil:");
            for (int z = rb.Zmin; z <= rb.Zmax; z++) {
                FrameGui.logger.log(Level.INFO, "----------------------------------------------------------------------------");
                FrameGui.logger.log(Level.INFO, "----------------------------------------------------------------------------");
                System.out.printf("z%d    ", z - rb.Zmin + 1);
                for (int y = rb.Ymin; y <= rb.Ymax; y++) {
                    System.out.printf("     y%d", (y - rb.Ymin) + 1);
                }
                FrameGui.logger.log(Level.INFO, "");
                FrameGui.logger.log(Level.INFO, "----------------------------------------------------------------------------");

                for (int x = rb.Xmin; x <= rb.Xmax; x++) {
                    System.out.printf("x%2d |", x - rb.Xmin + 1);

                    for (int y = rb.Ymin; y <= rb.Ymax; y++) {
                        System.out.printf("   %2d", rawZYXDiscret.getRealX_Y_Z(x, y, z));
                    }
                    FrameGui.logger.log(Level.INFO, "");
                }
                FrameGui.logger.log(Level.INFO, "\n");
            }*/
            //</editor-fold>
            int levelMax = (int) Math.ceil(rawZYXDiscret.maxValue);

            FrameGui.LOGGER.log(Level.INFO, "GLZLM levelMax = {0}", levelMax);
            TextureGLZLMSparse texGLZLMSparse = new TextureGLZLMSparse();
            
            //<editor-fold defaultstate="collapsed" desc="calculation of H">
            List<Future<Long>> results = new ArrayList<>();

            ExecutorService executor = Executors.newFixedThreadPool(nbCores);
            for (int iGrayLevel = 1; iGrayLevel <= levelMax; iGrayLevel++) {
                Callable<Long> callable = new TextureGLZLMSubRoutine(texGLZLMSparse, rb, rawZYXDiscret, maskRaw, iGrayLevel, messageGUI);
                results.add(executor.submit(callable));
            }
            
            executor.shutdown();
            try {
                executor.awaitTermination(5, TimeUnit.HOURS);
            } catch (InterruptedException ex) {
                FrameGui.LOGGER.log(Level.SEVERE, null, ex);
            }
            //</editor-fold> 
            //</editor-fold> 

            //<editor-fold defaultstate="collapsed" desc="save GLZLM.csv file">
            if (csvGLZLMSave) {
                FileOutputStream fileout = null;
                try {
                    long time = System.currentTimeMillis();
                    fileout = new FileOutputStream(outputFilename + "_GLZLM_" + time + ".csv");
                    try ( BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(fileout))) {

                        writer.write("GLZLM");
                        writer.newLine();
                        writer.write("(i j) see TextureUserGuide.pdf - 4.1");
                        writer.newLine();
                        writer.newLine();

                        int iLevel, iZone;

                        //<editor-fold defaultstate="collapsed" desc="get maxNbZones">
                        Map.Entry entryTaille, entryLevel;
                        HashMap hashtableTaille;
                        Iterator enLevel, enTaille;
                        ArrayList<Integer> listZone = new ArrayList<>();

                        enLevel = texGLZLMSparse.hashtableRow.entrySet().iterator();
                        while (enLevel.hasNext()) { // level
                            entryLevel = (Map.Entry) enLevel.next();

                            hashtableTaille = (HashMap) entryLevel.getValue();
                            enTaille = hashtableTaille.entrySet().iterator();
                            while (enTaille.hasNext()) { // zone
                                entryTaille = (Map.Entry) enTaille.next();
                                iZone = (Integer) entryTaille.getKey();
                                listZone.add(iZone);
                            }
                        }
                        Collections.sort(listZone);
                        int maxNbZones = listZone.get(listZone.size() - 1);
                        //</editor-fold>

                        writer.write("Row:Intensity(i):\t");
                        writer.newLine();

                        writer.write("Column:ZoneSize(j):\t");
                        writer.newLine();

                        // zone titles
                        writer.write("\t");
                        for (iZone = 0; iZone <= maxNbZones; iZone++) {
                            writer.write("j" + (iZone + 1) + "\t");
                        }
                        writer.newLine();

                        // write H
                        for (iLevel = 0; iLevel <= levelMax; iLevel++) {
                            writer.write("i" + (iLevel + 1) + "\t");

                            for (iZone = 0; iZone <= maxNbZones; iZone++) {
                                writer.write(texGLZLMSparse.get(iLevel, iZone) + "\t");
                            }
                            writer.newLine();
                        }
                        writer.newLine();

                    }
                    FrameGui.LOGGER.log(Level.INFO, "Save GLZLM.csv file: " + fileout);
                    MessageGui messageGUIOK = new MessageGui(
                            "csv File [ OK ]",
                            "GLZLM.csv",
                            MessageGui.OK_MESSAGE);
                } catch (IOException ex) {
                    FrameGui.LOGGER.log(Level.SEVERE,
                            null,
                            ex);
                } finally {
                    try {
                        if (fileout != null) {
                            fileout.close();
                        }
                    } catch (IOException ex) {
                        FrameGui.LOGGER.log(Level.SEVERE,
                                null, ex);
                    }
                }
            }
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="result of TexGLZLMSubRoutine">
            long teta = 0;
            for (Future<Long> future : results) {
                try {
                    try {
                        Long tetaOfSubRoutine = future.get();
                        teta += tetaOfSubRoutine;
                    } catch (InterruptedException ex) {
                        // Logger.getLogger(TexGLZLM.class.getName()).log(Level.SEVERE, null, ex);
                        FrameGui.LOGGER.log(Level.SEVERE, "GLZLM Processing ... [  OutOfMemoryError  ]");
                        outOfMemory = true;
                        return;
                    }

                } catch (ExecutionException ex) {
                    //Logger.getLogger(TexGLZLM.class.getName()).log(Level.SEVERE, null, ex);
                    FrameGui.LOGGER.log(Level.SEVERE, "GLZLM Processing ... [  OutOfMemoryError  ]");
                    outOfMemory = true;
                    return;
                }
            }
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="SZE LZE LGZE HGZE SZLGE SZHGE LZLGE LZHGE processing">
            SZE = 0;
            LZE = 0;
            LGZE = 0;
            HGZE = 0;
            SZLGE = 0;
            SZHGE = 0;
            LZLGE = 0;
            LZHGE = 0;
            {
                double value, iTaille2, iLevel2;
                int iLevel, iTaille;
                Map.Entry entryTaille, entryLevel;
                HashMap hashtableTaille;
                Iterator enLevel, enTaille;

                enLevel = texGLZLMSparse.hashtableRow.entrySet().iterator();
                //FrameGui.logger.log(Level.INFO, "enLevel : " + H.hashtableLevel.size());
                while (enLevel.hasNext()) {
                    entryLevel = (Map.Entry) enLevel.next();
                    iLevel = (Integer) entryLevel.getKey();

                    hashtableTaille = (HashMap) entryLevel.getValue();
                    //FrameGui.logger.log(Level.INFO, "--enTaille : " + hashtableTaille.size());
                    enTaille = hashtableTaille.entrySet().iterator();
                    while (enTaille.hasNext()) {

                        entryTaille = (Map.Entry) enTaille.next();

                        iTaille = (Integer) entryTaille.getKey();
                        if (iTaille > 0) {
                            value = (short) entryTaille.getValue();

                            iTaille2 = iTaille * iTaille;
                            iLevel2 = iLevel * iLevel;
                            //FrameGui.LOGGER.log(Level.INFO, "--iLevel : " + iLevel + " iTaille=" + iTaille + " value="+value + " iTaille2=" + iTaille2 + " iLevel="+iLevel2);
                            //System.out.println("--iLevel : " + iLevel + " iTaille=" + iTaille + " value="+value);

                            SZE += (value / iTaille2);
                            LZE += (value * iTaille2);
                            LGZE += (value / iLevel2);
                            HGZE += (value * iLevel2);
                            SZLGE += (value / (iLevel2 * iTaille2));
                            SZHGE += (value * iLevel2 / iTaille2);
                            LZLGE += (value * iTaille2 / iLevel2);
                            LZHGE += (value * iTaille2 * iLevel2);
                        }
                    }
                }
            }
            SZE = SZE / teta;
            LZE = LZE / teta;
            LGZE = LGZE / teta;
            HGZE = HGZE / teta;
            SZLGE = SZLGE / teta;
            SZHGE = SZHGE / teta;
            LZLGE = LZLGE / teta;
            LZHGE = LZHGE / teta;
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="Gray Level Non-uniformity GLNU processing">
            GLNU = 0;
            {
                short value;
                int iSize;
                Map.Entry entrySize, entryLevel;
                HashMap hashtableSize;
                Iterator enLevel, enSize;

                enLevel = texGLZLMSparse.hashtableRow.entrySet().iterator();
                //FrameGui.logger.log(Level.INFO, "enLevel : " + H.hashtableLevel.size());
                while (enLevel.hasNext()) {
                    entryLevel = (Map.Entry) enLevel.next();
                    //int iLevel = (Integer) entryLevel.getKey();

                    hashtableSize = (HashMap) entryLevel.getValue();
                    //FrameGui.logger.log(Level.INFO, "--enTaille : " + hashtableTaille.size());
                    value = 0;

                    enSize = hashtableSize.entrySet().iterator();
                    while (enSize.hasNext()) {

                        entrySize = (Map.Entry) enSize.next();

                        iSize = (Integer) entrySize.getKey();
                        if (iSize > 0) {
                            value += (short) entrySize.getValue();
                        }
                    }
                    GLNU += value * value;
                }
            }
            GLNU = GLNU / teta;
FrameGui.LOGGER.log(Level.INFO, "GLNU = " + GLNU);

            //</editor-fold>
            
            //<editor-fold defaultstate="collapsed" desc="H inverted">
            TextureGLZLMSparse Hinv = new TextureGLZLMSparse();
            {
                short value;
                int iLevel, iSize;

                Map.Entry entrySize, entryLevel;
                HashMap hashtableSize;
                Iterator enLevel, enSize;

                enLevel = texGLZLMSparse.hashtableRow.entrySet().iterator();

                while (enLevel.hasNext()) {
                    entryLevel = (Map.Entry) enLevel.next();
                    iLevel = (Integer) entryLevel.getKey();

                    hashtableSize = (HashMap) entryLevel.getValue();

                    enSize = hashtableSize.entrySet().iterator();
                    while (enSize.hasNext()) {

                        entrySize = (Map.Entry) enSize.next();

                        iSize = (Integer) entrySize.getKey();
                        value = (short) entrySize.getValue();
                        Hinv.set(iSize, iLevel, value);
                    }
                }
            }
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="Zone Length Non-uniformity ZLNU processing (no Optimized)">
            /*    ZLNU = 0;
             for (int iTaille = 1; iTaille <= iTailleMax; iTaille++) {
             int tmp = 0;
             for (int iLevel = 1; iLevel <= levelMax; iLevel++) {
             short value = H.get(iLevel, iTaille);
             if (value > 0) {
             tmp += value;
             }
             }
             ZLNU += tmp * tmp;
             }
             ZLNU = ZLNU / teta;
             */
            //</editor-fold>
            //<editor-fold defaultstate="collapsed" desc="Zone Length Non-uniformity ZLNU processing (Optimized)">
            ZLNU = 0;
            {
                // int iLevel;
                int iSize, tmp;
                Map.Entry entrySize, entryLevel;
                HashMap hashtableLevel;
                Iterator enLevel, enSize;

                enSize = Hinv.hashtableRow.entrySet().iterator();
                while (enSize.hasNext()) {
                    entrySize = (Map.Entry) enSize.next();
                    iSize = (Integer) entrySize.getKey();

                    if (iSize > 0) {
                        hashtableLevel = (HashMap) entrySize.getValue();
                        enLevel = hashtableLevel.entrySet().iterator();

                        tmp = 0;
                        while (enLevel.hasNext()) {

                            entryLevel = (Map.Entry) enLevel.next();
                            //iLevel = (Integer) entrySize.getKey();
                            tmp += (short) entryLevel.getValue();
                        }
                        ZLNU += tmp * tmp;
                    }
                }
                ZLNU = ZLNU / teta;
            }
            //</editor-fold>

            //<editor-fold defaultstate="collapsed" desc="ZonePercentage processing (No Optimized)">
            /*ZP = 0;
             for (int iTaille = 1; iTaille <= iTailleMax; iTaille++) {
             int tmp = 0;
             for (int iLevel = 1; iLevel <= levelMax; iLevel++) {
             tmp += iTaille * H.get(iLevel, iTaille);
             }
             ZP += tmp;
             }
             ZP = teta / ZP;*/
            //</editor-fold>
            //<editor-fold defaultstate="collapsed" desc="ZonePercentage processing (No Optimized)">
            // be careful: iLevel = iTaille et iTaille = iLevel (RawZYXSparce inversée)
            ZP = 0;
            {
                int iSize, tmp;
                Map.Entry entrySize, entryLevel;
                HashMap hashtableLevel;
                Iterator enLevel, enSize;

                enSize = Hinv.hashtableRow.entrySet().iterator();
                while (enSize.hasNext()) {
                    entrySize = (Map.Entry) enSize.next();
                    iSize = (Integer) entrySize.getKey();

                    if (iSize > 0) {
                        hashtableLevel = (HashMap) entrySize.getValue();
                        enLevel = hashtableLevel.entrySet().iterator();

                        tmp = 0;
                        while (enLevel.hasNext()) {

                            entryLevel = (Map.Entry) enLevel.next();
                            //iLevel = (Integer) entrySize.getKey();
                            tmp += iSize * (short) entryLevel.getValue();
                        }
                        ZP += tmp;
                    }
                }
                ZP = teta / ZP;
            }
            //</editor-fold>

        } catch (OutOfMemoryError ex) {
            outOfMemory = true;
            FrameGui.LOGGER.log(Level.SEVERE, "GLZLM Processing ... [  OutOfMemoryError  ]");
            SZE = 0;
            LZE = 0;
            LGZE = 0;
            HGZE = 0;
            SZLGE = 0;
            SZHGE = 0;
            LZLGE = 0;
            LZHGE = 0;
            GLNU = 0;
            ZLNU = 0;
            ZP = 0;
            return;
        }

        messageGUI.incProgress();
        FrameGui.LOGGER.log(Level.INFO, "GLZLM Processing ... [ end ]");
        long end_time = System.currentTimeMillis();
        long difference = end_time - start_time;
        FrameGui.LOGGER.log(Level.INFO, "Time calculation of GLZLM proceding: {0} sec", (difference / 1_000d));

    }

    /*
     @Override
     public String toString() {
     return "\nSecond Order (GLZLM matrix):"
     + "\n  SZE:\t\t" + getSZE()
     + "\n  LZE:\t\t" + getLZE()
     + "\n  LGZE:\t\t" + getLGZE()
     + "\n  HGzE:\t\t" + getHGZE()
     + "\n  SZLGE:\t\t" + getSZLGE()
     + "\n  SZHGE:\t\t" + getSZHGE()
     + "\n  LZLGE:\t\t" + getLZLGE()
     + "\n  LZHGE:\t\t" + getLZHGE()
     + "\n  GLNU:\t\t" + getGLNU()
     + "\n  ZLNU:\t\t" + getZLNU()
     + "\n  ZP:\t\t" + getZP();
     }*/
    public double getGLNU() {
        return GLNU;
    }

    public double getHGZE() {
        return HGZE;
    }

    public double getLGZE() {
        return LGZE;
    }

    public double getLZE() {
        return LZE;
    }

    public double getLZHGE() {
        return LZHGE;
    }

    public double getLZLGE() {
        return LZLGE;
    }

    public double getSZE() {
        return SZE;
    }

    public double getSZHGE() {
        return SZHGE;
    }

    public double getSZLGE() {
        return SZLGE;
    }

    public double getZLNU() {
        return ZLNU;
    }

    public double getZP() {
        return ZP;
    }

    public void setIsTooSmall() {
       SZE =  Double.NaN;
       LZE =  Double.NaN;
       LGZE =  Double.NaN;
       HGZE =  Double.NaN;
       SZLGE =  Double.NaN;
       SZHGE =  Double.NaN;
       LZLGE =  Double.NaN;
       LZHGE =  Double.NaN;
       GLNU =  Double.NaN;
       ZLNU =  Double.NaN;
       ZP =  Double.NaN;
    }
    
    /**
     * true if outOfMemory processing
     */
    public boolean outOfMemory;

    private final RawSparseZYX maskRaw;
    private final CoorRange rb;

    /**
     * discrete image
     */
    private final RawTZYX rawZYXDiscret;

    /**
     * Short Zone Emphasis SZE
     */
    private double SZE;
    /**
     * Long Zone Emphasis LZE
     */
    private double LZE;
    /**
     * Low Gray-Level Zone Emphasis LGZE
     */
    private double LGZE;
    /**
     * High Gray-Level Zone Emphasis HGZE
     */
    private double HGZE;
    /**
     * Short Zone Low Gray-Level Emphasis SZLGE
     */
    private double SZLGE;
    /**
     * Short Zone High Gray-Level Emphasis SZHGE
     */
    private double SZHGE;
    /**
     * Long Zone Low Gray-Level Emphasis LZLGE
     */
    private double LZLGE;
    /**
     * Long Zone High Gray-Level Emphasis LZHGE
     */
    private double LZHGE;
    /**
     * Gray-Level Non-Uniformity GLNU
     */
    private double GLNU;
    /**
     * Zone Length Non-Uniformity ZLNU
     */
    private double ZLNU;
    /**
     * Zone Percentage
     */
    private double ZP;

    private final MessageGui messageGUI;
    private final int nbCores;

    private final String outputFilename;
    private final boolean csvGLZLMSave;

    
}
