/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.textureGLZLM;

import bardp.img.coor.CoorRange;
import bardp.img.coor.Coor;
import bardp.raw.RawSparseZYX;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class TextureGLZLMCluster3D {

    /**
     * label of clusters from [2 .. n]
     *
     * @param rawZYXSparce binary [0, 1]
     * @param rangeBounds to define
     */
    public TextureGLZLMCluster3D(
            final RawSparseZYX rawZYXSparce,
            final CoorRange rangeBounds
    ) {
        this.rawZYXSparce = rawZYXSparce;
        this.rangeBounds = rangeBounds;
    }

    /**
     * be careful, return the maxID of cluster and not the number of clusters.
     * <br>ID = 2 -> first index of cluster
     *
     * @return cluster id [2..num]
     */
    public int run() {

        int rbMaxFloorZ = rangeBounds.max.getZ();
        int rbMinFloorZ = rangeBounds.min.getZ();
        int rbMaxFloorY = rangeBounds.max.getY();
        int rbMinFloorY = rangeBounds.min.getY();
        int rbMaxFloorX = rangeBounds.max.getX();
        int rbMinFloorX = rangeBounds.min.getX();

        nbVoxels = new HashMap<>();

        short[][][] clusterMap = new short
                [rbMaxFloorZ - rbMinFloorZ + 1 + 2] // +1=normal   +2=border security
                [rbMaxFloorY - rbMinFloorY + 1 + 2]
                [rbMaxFloorX - rbMinFloorX + 1 + 2];

        short id = 2; // be careful: 2 is important (and not 1 or 0)

        // change format of mask rawZYXSparce -> clusterMap
        for (Coor coor : rawZYXSparce.getCoorList()) {
            clusterMap[(coor.getZ() - rbMinFloorZ + 1)]
                    [(coor.getY() - rbMinFloorY + 1)]
                    [(coor.getX() - rbMinFloorX + 1)] = 1; // initialization of clusterId
        }

        int[][] neighbourMatrix;
        short cell;
        boolean neighbourFound;
        short neighbourId;
        int i;
        int z, y, x;
        int neighboursLength;
        short[] neighbours;
        int val;

        for (z = rbMinFloorZ; z <= rbMaxFloorZ; z++) {

            // 2D or 3D matrix neighbour managment
            if (z == rbMinFloorZ) {
                neighbourMatrix = NEIGHBOURS2D;
            } else {
                neighbourMatrix = NEIGHBOURS3D;
            }
            neighboursLength = neighbourMatrix[0].length;
            neighbours = new short[neighboursLength];

            for (y = rbMinFloorY; y <= rbMaxFloorY; y++) {
                for (x = rbMinFloorX; x <= rbMaxFloorX; x++) {
                    if (clusterMap[z - rbMinFloorZ + 1]
                            [y - rbMinFloorY + 1]
                            [x - rbMinFloorX + 1] > 0) {

                        // init neighbours
                        neighbourId = Short.MAX_VALUE;
                        for (i = 0; i < neighbours.length; i++) {
                            neighbours[i] = 0;
                        }
                        neighbourFound = false;

                        // reserch of neighbours
                        for (i = 0; i < neighboursLength; i++) {
                            cell = clusterMap
                                    [z - rbMinFloorZ + 1 + neighbourMatrix[2][i]]
                                    [y - rbMinFloorY + 1 + neighbourMatrix[1][i]]
                                    [x - rbMinFloorX + 1 + neighbourMatrix[0][i]];
                            if (cell > 1) {
                                neighbourFound = true;
                                neighbours[i] = cell;
                                if (cell < neighbourId) {
                                    neighbourId = cell;
                                }
                            }
                        }

                        if (neighbourFound) {
                            clusterMap
                                    [z - rbMinFloorZ + 1]
                                    [y - rbMinFloorY + 1]
                                    [x - rbMinFloorX + 1] = neighbourId;

                            inc(neighbourId);

                            for (i = 0; i < neighboursLength; i++) {
                                if ((neighbours[i] > 0) && (neighbours[i] != neighbourId)) {
                                    if (nbVoxels.containsKey(neighbours[i])) {
                                        val = nbVoxels.get(neighbours[i]);
                                        nbVoxels.remove(neighbours[i]);
                                        if (nbVoxels.containsKey(neighbourId)) {
                                            nbVoxels.put(neighbourId, nbVoxels.get(neighbourId) + val);
                                        } else {
                                            nbVoxels.put(neighbourId, val);
                                        }
                                    }
                                    clusterMap = changeId(clusterMap, neighbourId, neighbours[i]);
                                }
                            }
                        } else {
                            clusterMap
                                    [z - rbMinFloorZ + 1]
                                    [y - rbMinFloorY + 1]
                                    [x - rbMinFloorX + 1] = id;
                            inc(id);
                            id++;

                        }
                    }
                }
            }
        }

        return id; // clusterId -> important for TextGLZLM
    }

    public HashMap<Short, Integer> getNbVoxels() {
        return nbVoxels;
    }
 

    /**
     * increase of 1 at id greyLevel
     * <br> if id greyLevel not exist, create it with 1 of value count
     * @param id 
     */
    private void inc(
            final short id
    ) {
        if (nbVoxels.containsKey(id)) {
            nbVoxels.put(id, nbVoxels.get(id) + 1);
        } else {
            nbVoxels.put(id, 1);
        }
    }

    private short[][][] changeId(
            final short[][][] t,
            final short newID,
            final short oldID
    ) {
        int nz = t.length;
        int ny = t[0].length;
        int nx = t[0][0].length;
        int z, y, x;
        for (z = 0; z < nz; z++) {
            for (y = 0; y < ny; y++) {
                for (x = 0; x < nx; x++) {
                    if (t[z][y][x] == oldID) {
                        t[z][y][x] = newID;
                    }
                }
            }
        }
        return t;
    }

    private final static int[][] NEIGHBOURS2D = {
        {-1, -1, 0, 1, 1, 1, 0, -1}, //x
        {0, -1, -1, -1, 0, 1, 1, 1}, //y
        {0, 0, 0, 0, 0, 0, 0, 0} //z
    };

    private final static int[][] NEIGHBOURS3D = {
        // direc  1   2   3   4   5   6   7   8   9  10  11  12 13  14 15 16  17
        {-1, -1, 0, 1, 1, 1, -1, 0, -1, 0, -1, 0, 1, 1, 1, 0, -1}, //x
        {0, -1, -1, -1, 0, 1, -1, -1, 0, 0, 1, 1, 0, -1, 1, 1, 1}, //y
        {0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0} //z
    };

    private HashMap<Short, Integer> nbVoxels;
    private final RawSparseZYX rawZYXSparce;
    private final CoorRange rangeBounds;
}
