/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.textureNGLDM;

import bardp.gui.FrameGui;
import bardp.img.coor.CoorRange;
import bardp.raw.RawTZYX;
import bardp.raw.RawSparseZYX;
import bardp.utility.MessageGui;
import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.logging.Level;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: Neighborhood Gray-Level Different Matrix</p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class TextureNGLDM implements Runnable {

   /**
    * processing the Neighborhood Gray-Level Different Matrix (NGLDM)
    * <p>
    * @param outputFilename
    * @param csvNGLDMSave
    * @param rawZYXDiscret discretized rawZYX (unit wanted: SUV)
    * @param rb
    * @param maskRaw
    * @param messageGUI
    */
   public TextureNGLDM(
           String outputFilename,
           boolean csvNGLDMSave,
           RawTZYX rawZYXDiscret,
           RawSparseZYX maskRaw,
           CoorRange rb,
           MessageGui messageGUI
   ) {
      this.outputFilename = outputFilename;
      this.csvNGLDMSave = csvNGLDMSave;
      this.rawZYXDiscrete = rawZYXDiscret;
      this.maskRaw = maskRaw;
      this.rb = rb;
      this.messageGUI = messageGUI;
   }

   @Override
   public void run() {

      long start_time = System.currentTimeMillis();
      FrameGui.LOGGER.log(Level.INFO, "NGLDM Processing ... [ start ]");

      //<editor-fold defaultstate="collapsed" desc="NGLDM core processing">
      int rbMaxFloorZ = rb.max.getZ();
      int rbMaxFloorY = rb.max.getY();
      int rbMaxFloorX = rb.max.getX();
      
      int rbMinFloorZ = rb.min.getZ();
      int rbMinFloorY = rb.min.getY();
      int rbMinFloorX = rb.min.getX();

      NGLDM = new double
              [rbMaxFloorX - rbMinFloorX + 1]
              [rbMaxFloorY - rbMinFloorY + 1]
              [rbMaxFloorZ - rbMinFloorZ + 1];

      int it = 0;
      
      int iz, iy, ix;
      int dz, dy, dx;
      double m, n;
      for (iy = rbMinFloorY; iy <= rbMaxFloorY; iy++) {
         for (ix = rbMinFloorX; ix <= rbMaxFloorX; ix++) {
            for (iz = rbMinFloorZ; iz <= rbMaxFloorZ; iz++) {

               if (maskRaw.getXYZ(ix, iy, iz) > 0) {
                  m = 0; // sum of neighborhoods
                  n = 0; // count of neighborhoods

                  for (dy = -1; dy <= +1; dy++) {
                     for (dx = -1; dx <= +1; dx++) {
                        for (dz = -1; dz <= +1; dz++) {

                           // center voxel only if it isn't in sum
                           if (((dx != 0) || (dy != 0)) || (dz != 0)) {
                              // only voxels inclued in ROI
                              if (maskRaw.getZYX(iz + dz, iy + dy, ix + dx) > 0) {
                                 m += rawZYXDiscrete.getTZYX(it, iz + dz, iy + dy, ix + dx);
                                 n++;
                              }
                           }
                        }
                     }
                  }
                  m /= n;
                  NGLDM[ix - rbMinFloorX][iy - rbMinFloorY][iz - rbMinFloorZ] = m;
               }
            }
         }
      }
      //</editor-fold>

      //<editor-fold defaultstate="collapsed" desc="Debug : Display NGLDM matrix">
      if (FrameGui.DEBUG) {
        System.out.println("_______________ NGLDM ___________________");
        for (iz = rbMinFloorZ; iz <= rbMaxFloorZ; iz++) {
            System.out.println("z=" + iz);
            for (iy = rbMinFloorY; iy <= rbMaxFloorY; iy++) {
                for (ix = rbMinFloorX; ix <= rbMaxFloorX; ix++) {
                    System.out.printf(" %3.4f", NGLDM[ix - rbMinFloorX][iy - rbMinFloorY][iz - rbMinFloorZ]);
                }
                System.out.println("");
            }
        }
      }
      //</editor-fold>
      
      //<editor-fold defaultstate="collapsed" desc="frequence / rang">
      double[] s = new double[(int) rawZYXDiscrete.maxValue + 1];
      double[] p = new double[(int) rawZYXDiscrete.maxValue + 1];

      int grayValue;
      for (iy = rbMinFloorY; iy <= rbMaxFloorY; iy++) {
         for (ix = rbMinFloorX; ix <= rbMaxFloorX; ix++) {
            for (iz = rbMinFloorZ; iz <= rbMaxFloorZ; iz++) {
               grayValue = Math.round(rawZYXDiscrete.getTZYX(it, iz, iy, ix));
               if (grayValue > 0) {
                  s[grayValue] += Math.abs(grayValue - NGLDM[ix - rbMinFloorX][iy - rbMinFloorY][iz - rbMinFloorZ]);
                  p[grayValue]++;
               }
            }
         }
      }

      int i, j;
      double nTotal = 0;
      for (i = 0; i < p.length; i++) {
         nTotal += p[i];
      }
      for (i = 0; i < p.length; i++) {
         p[i] /= nTotal;
      }
      //</editor-fold>

      //<editor-fold defaultstate="collapsed" desc="reduct of p and s array with remove all 0 values">
      ArrayList<Integer> _ii = new ArrayList<>();
      ArrayList<Double> _pp = new ArrayList<>();
      for (i = 0; i < p.length; i++) {
         if (p[i] > 0) {
            _pp.add(p[i]);
            _ii.add(i);
         }
      }
      p = new double[_pp.size()];
      int[] indexGrey = new int[_pp.size()];
      for (i = 0; i < _pp.size(); i++) {
         p[i] = _pp.get(i);
         indexGrey[i] = _ii.get(i);
      }
      _pp = null;
      _ii = null;
      
      ArrayList<Double> _ss = new ArrayList<>();
      for (i = 0; i < s.length; i++) {
         if (s[i] > 0) {
            _ss.add(s[i]);
         }
      }

      s = new double[_ss.size()];
      for (i = 0; i < _ss.size(); i++) {
         s[i] = _ss.get(i);
      }
      _ss = null;
      //</editor-fold>

      //<editor-fold defaultstate="collapsed" desc="save NGLDM.csv file">
      if (csvNGLDMSave) {
         FileOutputStream fileout = null;
         try {
            long time = System.currentTimeMillis();
            fileout = new FileOutputStream(outputFilename + "_NGLDM_" + time + ".csv");
            try (BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(fileout))) {

               writer.write("NGLDM");
               writer.newLine();
               writer.write("TextureUserGuide.pdf - 2.1p20");
               writer.newLine();
               writer.newLine();

               writer.write("i\tD([i][1])\tD([i][2])");
               writer.newLine();
               for (i = 0; i < p.length; i++) {
                  writer.write(indexGrey[i] + "\t" + p[i] + "\t" + s[i] + "\t");
                  writer.newLine();
               }
            }
            FrameGui.LOGGER.log(Level.INFO, "Save NGLDM.csv file: {0}", fileout);
            MessageGui messageGUI = new MessageGui(
                    "csv File [ OK ]",
                    "NGLDM.csv",
                    MessageGui.OK_MESSAGE);
         } catch (IOException ex) {
            FrameGui.LOGGER.log(Level.SEVERE,
                    null,
                    ex);
         } finally {
            try {
               if (fileout != null) {
                  fileout.close();
               }
            } catch (IOException ex) {
               FrameGui.LOGGER.log(Level.SEVERE,
                       null, ex);
            }
         }
      }
      //</editor-fold>

      //<editor-fold defaultstate="collapsed" desc="Coarseness processing">
      {
         double sum = 0;
         for (i = 0; i < p.length; i++) {
            sum += p[i] * s[i];
         }
         if (sum == 0) {
            coarseness = 1000000;
         } else {
            coarseness = 1d / sum;
         }
         FrameGui.LOGGER.log(Level.INFO, "NGLDM_Coarseness={0}", coarseness);
      }
      //</editor-fold>

      //<editor-fold defaultstate="collapsed" desc="Contrast processing">
      {
         double temp = 0;
         double part = 0;
         double i_j;
         int ii, jj;
         
         for (i = 0; i < p.length; i++) {

            part += s[i];

            for (j = 0; j < p.length; j++) {
               ii = indexGrey[i];
               jj = indexGrey[j];
               
               i_j = ii - jj;
               temp += p[i] * p[j] * i_j * i_j;
            }
         }

         contrast = (temp / (p.length * (p.length - 1))) * (part / (nTotal)); // not nTotal^2
         FrameGui.LOGGER.log(Level.INFO, "NGLDM_Contrast={0}", contrast);
      }
      //</editor-fold>

      //<editor-fold defaultstate="collapsed" desc="Busyness processing">
      {
         double temp = 0;
         double part = 0;
         int ii, jj;
         
         for (i = 0; i < p.length; i++) {
            part += p[i] * s[i];
            ii = indexGrey[i];
            
            for (j = 0; j < p.length; j++) {
               jj = indexGrey[j];
               temp += Math.abs(ii*p[i] - jj*p[j]); // corrected with indexGrey[i]
            }
         }

         if (temp == 0) {
            busyness = 0;
         } else {
            busyness = part / temp;
         }
         FrameGui.LOGGER.log(Level.INFO, "NGLDM_Busyness={0}", busyness);
      }

      //</editor-fold>
      FrameGui.LOGGER.log(Level.INFO, "NGLDM Processing ... [ end ]");
      messageGUI.incProgress();

      long end_time = System.currentTimeMillis();
      long difference = end_time - start_time;
      FrameGui.LOGGER.log(Level.INFO, "NGLDM: Time calculation: {0} sec", (difference / 1_000d));
   }

   public double getBusyness() {
      return busyness;
   }

   public double getCoarseness() {
      return coarseness;
   }

   public double getComplexity() {
      return complexity;
   }

   public double getContrast() {
      return contrast;
   }

   public double getTextureStrength() {
      return textureStrength;
   }

   @Override
   public String toString() {
      return "\nSecond Order (NGLDM matrix):"
              + "\n  Coarseness:\t\t" + getCoarseness()
              + "\n  Contrast:\t\t" + getContrast()
              + "\n  Busyness:\t\t" + getBusyness();
   }

   
   public void setIsTooSmall() {
        coarseness = Double.NaN;
        contrast = Double.NaN;
        busyness = Double.NaN;
        complexity = Double.NaN;
        textureStrength = Double.NaN;
    }
   
   
   /**
    * discrete image
    */
   private final RawTZYX rawZYXDiscrete;
   
   
   private final RawSparseZYX maskRaw;
   private final CoorRange rb;

   private double[][][] NGLDM;

   private double coarseness;

   private double contrast;

   private double busyness;

   private double complexity;

   private double textureStrength;

   private final MessageGui messageGUI;

   private final String outputFilename;
   private final boolean csvNGLDMSave;

    
}
