/*
 * USE OF THE LIFEx SOFTWARE IS SUBJECT TO THE FOLLOWING TERMS AND CONDITIONS. 
 * IF YOU DO NOT ACCEPT THESE TERMS AND CONDITIONS, DO NOT DOWNLOAD THE SOFTWARE. 
 * NON-PROFESSIONAL USERS ARE NOTIFIED THAT THE SOFTWARE AND THE RISKS INHERENT TO ITS USE,
 * MODIFICATION AND/OR DEVELOPMENT AND REPRODUCTION ARE NOT COVERED BY ANY WARRANTY ON THE PART OF THE CEA. 
 * IT IS HIGHLY RECOMMENDED THAT YOU ONLY LOAD AND/OR USE THE SOFTWARE IF YOU HAVE SPECIALIZED IN-DEPTH INFORMATION TECHNOLOGY KNOW-HOW.
 * THE COMMISSARIAT A L’ENERGIE ATOMIQUE ET AUX ENERGIES ALTERNATIVES (FRENCH ATOMIC ENERGY AGENCY), 
 * a public scientific, technical and industrial establishment, 
 * having its registered office at 25 rue Leblanc – bâtiment le Ponant D – 75015 Paris cedex, 
 * hereinafter referred to as "CEA", hereby licenses the right to use the LIFEx software as per the terms and conditions set out below.  
 */
package bardp.protocol.protocolTexture.traditional;

import bardp.gui.FrameGui;
import bardp.img.coor.Coor;
import bardp.img.coor.CoorList;
import bardp.raw.RawSparseZYX;
import bardp.raw.RawTZYX;
import bardp.roi.calc.RoiPeakModel;
import java.util.logging.Level;

/**
 * <p>
 * Title: Local Image Feature Extraction (LIFEx)</p>
 * <p>
 * Description: </p>
 * <p>
 * Copyright: Copyright (c) 2016</p>
 * <p>
 * Company: Inserm/CNRS/CEA</p>
 * <p>
 * @author Christophe Nioche, Inserm at lifexsoft.org
 * @version 6.0
 */
public class ConventionalPeak {

    /**
     * available parameters of ConvPeak class
     */
    public enum ConvPeakParams {

        /**
         * define search volume at 1 cm3
         */
        PEAK_VOL1CM3,
        /**
         * define search volume at 0.5 cm3
         */
        PEAK_VOL05CM3
    };

    /**
     *
     * @param rawZYXContinue
     * @param rawZYXSparse
     * @param convPeakParamsSearchVolume
     */
    public ConventionalPeak(
            final RawTZYX rawZYXContinue,
            final RawSparseZYX rawZYXSparse,
            final ConvPeakParams convPeakParamsSearchVolume
    ) {
        this.rawTZYXContinue = rawZYXContinue;
        this.rawZYXSparse = rawZYXSparse;
        this.convPeakParams = convPeakParamsSearchVolume;
    }

    public RoiPeakModel run() {
        long start_time = System.currentTimeMillis();

        RoiPeakModel peakModel = new RoiPeakModel();

        //<editor-fold defaultstate="collapsed" desc="coorListSolid3D = define vx coordinates of solid3D">
        float radiusVx = 0;
        double vol;
        int ii;

        double searchVolume
                = convPeakParams == ConvPeakParams.PEAK_VOL1CM3 ? 1_000 // mm3
                        : convPeakParams == ConvPeakParams.PEAK_VOL05CM3 ? 500 // mm3
                                : 1_000; // mm3
        do {
            radiusVx += 0.1f;
            coorListSolid3D = createSolid3D(radiusVx);
            vol = coorListSolid3D.size() * rawTZYXContinue.iopIppSp.spModel.getVoxelSpacing();
        } while (vol < searchVolume);
        
        FrameGui.LOGGER.log(
                Level.INFO,
                "vol = {0}    radiusVx = {1}    coorListSolid3D ={2}",
                new Object[]{
                    vol, radiusVx, coorListSolid3D.size()
                }
        );

        // recalcul of state before the last one
        CoorList coorListSolid3DBefore = createSolid3D(radiusVx - 0.1f - 0.1f);
        double volBefore = coorListSolid3DBefore.size() * rawTZYXContinue.iopIppSp.spModel.getVoxelSpacing();

        // find last vol to check volume difference
        if (Math.abs(vol - searchVolume) > Math.abs(volBefore - searchVolume)) {
            coorListSolid3D = coorListSolid3DBefore;
            vol = volBefore;
        }
        //</editor-fold>

        peakModel.peakDiscretizedVolumeSought = (float) vol;

        //<editor-fold defaultstate="collapsed" desc="define temporary variables">
        int iz, iy, ix;
        float value;
        float mean;
        float squareSum;
        float counts;
        boolean isOutside;
        //</editor-fold>

        int[][] coorListSphere = new int[coorListSolid3D.size()][3];
        float coorListSphereSizeFloat = (float) coorListSolid3D.size();
        int coorListSphereEndSizeInt = (int) coorListSphereSizeFloat - 1;

        final float[][] rawZYXContinue = rawTZYXContinue.getZYX(rawTZYXContinue.t);
        final int nx = rawTZYXContinue.nx;

        // decrease time of execution with linear array
        short[][][] rawZYX = rawZYXSparse.convZYXInDenseFormatShortShortShort();

        for (iz = 0; iz < rawZYXSparse.nz; iz++) {
            for (iy = 0; iy < rawZYXSparse.ny; iy++) {
                for (ix = 0; ix < rawZYXSparse.nx; ix++) {

                    // continue if voxel is ouside of ROI
                    if (rawZYX[iz][iy][ix] == 0) {
                        continue;
                    }

                    // get all coordinates of solid3D with center (z, y, x)
                    this.setCoorWithCenter(coorListSphere, iz, iy, ix);

                    mean = 0f;
                    squareSum = 0f;
                    isOutside = false;
                    //<editor-fold defaultstate="collapsed" desc="finding all solid3D coordinates included in rawZYXSparse: else boundEffect is true">
                    for (ii = coorListSphereEndSizeInt; ii >= 0; ii--) {

                        try {
                            if (rawZYX[coorListSphere[ii][0]][coorListSphere[ii][1]][coorListSphere[ii][2]] == 0) {
                                isOutside = true;
                                break;
                            }
                        } catch (ArrayIndexOutOfBoundsException ignore) {
                            isOutside = true;
                            break;
                        }

                        value = rawZYXContinue[coorListSphere[ii][0]][coorListSphere[ii][1] * nx
                                + coorListSphere[ii][2]];

                        squareSum += value * value;
                        mean += value;
                    }

                    //</editor-fold>
                    if (isOutside) {
                        continue;
                    }

                    counts = coorListSphereSizeFloat;
                    mean /= counts;
                    //</editor-fold>

                    //<editor-fold defaultstate="collapsed" desc="keep mean in peakValue if necessary">
                    if (mean > peakModel.peakValue) {
                        peakModel.peakValue = mean;
                        peakModel.peakStdev = (float) Math.sqrt(squareSum / counts - mean * mean);
                        peakModel.peakCoorCenter.set(iz, iy, ix);
                        peakModel.peakTimeLeft = 0; // no define here
                    }
                }
            }
        }

        if (peakModel.peakValue == 0) {
            peakModel.peakValue = Float.NaN;
            peakModel.peakStdev = Float.NaN;
        }

        long end_time = System.currentTimeMillis();
        long difference = end_time - start_time;
        FrameGui.LOGGER.log(
                Level.INFO,
                "Time calculation of peak proceding: {0} sec", (difference / 1_000d)
        );
        return peakModel;
    }

    /**
     * create solid3D coordinate of 3D plain sphere with radius "radiusVx"
     *
     * @param radiusVx
     * @return
     */
    private CoorList createSolid3D(
            final float radiusVx
    ) {
        // isovoxel conversion
        double radiusZ = radiusVx / rawTZYXContinue.iopIppSp.spModel.getZSpacing();
        double radiusY = radiusVx / rawTZYXContinue.iopIppSp.spModel.getYSpacing();
        double radiusX = radiusVx / rawTZYXContinue.iopIppSp.spModel.getXSpacing();

        // round conversion for pixel status
        double radiusZRound = Math.round(radiusZ);
        double radiusYRound = Math.round(radiusY);
        double radiusXRound = Math.round(radiusX);

        double pixZ, pixY, pixX;
        double pow2X, pow2Y, pow2Z;

        final CoorList solid3D = new CoorList();

        for (pixZ = -radiusZRound; pixZ <= radiusZRound; pixZ++) {
            pow2Z = Math.pow(pixZ / radiusZ, 2);

            for (pixY = -radiusYRound; pixY <= radiusYRound; pixY++) {
                pow2Y = Math.pow(pixY / radiusY, 2);

                for (pixX = -radiusXRound; pixX <= radiusXRound; pixX++) {
                    pow2X = Math.pow(pixX / radiusX, 2);

                    if (Math.sqrt(pow2X + pow2Y + pow2Z) <= 1) {
                        solid3D.add(new Coor((int) pixZ, (int) pixY, (int) pixX));
                    }
                }
            }
        }
        return solid3D;
    }

    /**
     * setting of coorListSphere with (z,y,x) center point
     *
     * @param coorListSphere exist [ii][zyx] for result setting
     * @param cz z coordinate of center
     * @param cy y coordinate of center
     * @param cx x coordinate of center
     */
    private void setCoorWithCenter(
            final int[][] coorListSphere,
            final int cz,
            final int cy,
            final int cx
    ) {
        Coor pCircle3D;
        for (int ii = coorListSolid3D.size() - 1; ii >= 0; ii--) {
            pCircle3D = coorListSolid3D.get(ii);
            coorListSphere[ii][0] = cz + pCircle3D.getZ();
            coorListSphere[ii][1] = cy + pCircle3D.getY();
            coorListSphere[ii][2] = cx + pCircle3D.getX();
        }
    }

    private final RawTZYX rawTZYXContinue;
    private final RawSparseZYX rawZYXSparse;

    /**
     * peak search volume in cm3
     */
    private final ConvPeakParams convPeakParams;

    /**
     * coordinate list of solid3D
     */
    private CoorList coorListSolid3D = new CoorList();
}
